(*
** $Id: lauxlib.h,v 1.52 2002/08/08 20:08:41 roberto Exp $
** Auxiliary functions for building Lua libraries
** See Copyright Notice in lua.h
*)
(*
** Translated to pascal by Lavergne Thomas
** Notes :
**    - Pointers type was prefixed with 'P'
** Bugs report :
**    - thomas.lavergne@laposte.net
**   In french or in english
*)
unit lauxlib;

interface

uses
  Lua;

type
  luaL_reg = record
    name: PChar;
    func: lua_CFunction;
  end;
  PluaL_reg = ^luaL_reg;

procedure luaL_openlib(L: Plua_State; const lr: PluaL_reg; nup: Integer); cdecl;
procedure luaL_opennamedlib(L: Plua_State; const libname: PChar; const lr: PluaL_reg; nup: Integer); cdecl;
function luaL_callmeta(L: Plua_State; obj: Integer; const event: PChar): Integer; cdecl;
function luaL_typerror(L: Plua_State; narg: Integer; const tname: PChar): Integer; cdecl;
function luaL_argerror(L: Plua_State; numarg: Integer; const extramsg: PChar): Integer; cdecl;
function luaL_check_lstr(L: Plua_State; numArg: Integer; len: Psize_t): PChar; cdecl;
function luaL_opt_lstr(L: Plua_State; numArg: Integer; const def: PChar; len: Psize_t): PChar; cdecl;
function luaL_check_number(L: Plua_State; numArg: Integer): lua_Number; cdecl;
function luaL_opt_number(L: Plua_State; nArg: Integer; def: lua_Number): lua_Number; cdecl;

procedure luaL_check_stack(L: Plua_State; space: Integer; const msg: PChar); cdecl;
procedure luaL_check_type(L: Plua_State; narg, t: Integer); cdecl;
procedure luaL_check_any(L: Plua_State; narg: Integer); cdecl;

procedure luaL_where(L: Plua_State; level: Integer); cdecl;
function luaL_error(L: Plua_State; const fmt: PChar; argsup: array of pointer): Integer; cdecl;

function luaL_findstring(name: PChar; const list: PPChar): Integer; cdecl;

function luaL_ref(L: Plua_State; t: Integer): Integer; cdecl;
procedure luaL_unref(L: Plua_State; t, ref: Integer); cdecl;

function luaL_loadfile(L: Plua_State; const filename: PChar): Integer; cdecl;
function luaL_loadbuffer(L: Plua_State; const buff: PChar; size: Integer; const name: PChar): Integer; cdecl;

(*
** ===============================================================
** some useful macros
** ===============================================================
*)

procedure luaL_arg_check(L: Plua_State; cond: Boolean; numarg: Integer; extramsg: PChar);
function luaL_check_string(L: Plua_State; n: Integer): PChar;
function luaL_opt_string(L: Plua_State; n: Integer; d: PChar): PChar;
function luaL_check_int(L: Plua_State; n: Integer): Integer;
function luaL_check_long(L: Plua_State; n: Integer): LongInt;
function luaL_opt_int(L: Plua_State; n: Integer; d: Double): Integer;
function luaL_opt_long(L: Plua_State; n: Integer; d: Double): LongInt;

(*
** {======================================================
** Generic Buffer manipulation
** =======================================================
*)

const
  LUAL_BUFFERSIZE = 4096;

type
  luaL_Buffer = record
    p: PChar;
    level: Integer;
    L: Plua_State;
    buffer: array [0..LUAL_BUFFERSIZE - 1] of Char;
  end;
  PluaL_Buffer = ^luaL_Buffer;

procedure luaL_putchar(B: PluaL_Buffer; c: Char);

procedure luaL_addsize(B: PluaL_Buffer; n: Integer);

procedure luaL_buffinit(L: Plua_State ; B: PluaL_Buffer); cdecl;
function luaL_prepbuffer(B: PluaL_Buffer): PChar; cdecl;
procedure luaL_addlstring(B: PluaL_Buffer; const s: PChar; l: size_t); cdecl;
procedure luaL_addstring(B: PluaL_Buffer; const s: PChar); cdecl;
procedure luaL_addvalue(B: PluaL_Buffer); cdecl;
procedure luaL_pushresult(B: PluaL_Buffer); cdecl;

(*
** Compatibility macros
*)

procedure luaL_checktype(L: Plua_State; narg, t: Integer);
procedure luaL_checkany(L: Plua_State; narg: Integer);

function lua_dofile(L: Plua_State; const filename: PChar): Integer; cdecl;
function lua_dostring(L: Plua_State; const str: PChar): Integer; cdecl;
function lua_dobuffer(L: Plua_State; const buff: PChar; size: Integer; const name: PChar): Integer; cdecl;

implementation

procedure luaL_openlib(L: Plua_State; const lr: PluaL_reg; nup: Integer); cdecl; external LUA_LIB_NAME;
procedure luaL_opennamedlib(L: Plua_State; const libname: PChar; const lr: PluaL_reg; nup: Integer); cdecl; external LUA_LIB_NAME;
function luaL_callmeta(L: Plua_State; obj: Integer; const event: PChar): Integer; cdecl; external LUA_LIB_NAME;
function luaL_typerror(L: Plua_State; narg: Integer; const tname: PChar): Integer; cdecl; external LUA_LIB_NAME;
function luaL_argerror(L: Plua_State; numarg: Integer; const extramsg: PChar): Integer; cdecl; external LUA_LIB_NAME;
function luaL_check_lstr(L: Plua_State; numArg: Integer; len: Psize_t): PChar; cdecl; external LUA_LIB_NAME;
function luaL_opt_lstr(L: Plua_State; numArg: Integer; const def: PChar; len: Psize_t): PChar; cdecl; external LUA_LIB_NAME;
function luaL_check_number(L: Plua_State; numArg: Integer): lua_Number; cdecl; external LUA_LIB_NAME;
function luaL_opt_number(L: Plua_State; nArg: Integer; def: lua_Number): lua_Number; cdecl; external LUA_LIB_NAME;

procedure luaL_check_stack(L: Plua_State; space: Integer; const msg: PChar); cdecl; external LUA_LIB_NAME;
procedure luaL_check_type(L: Plua_State; narg, t: Integer); cdecl; external LUA_LIB_NAME;
procedure luaL_check_any(L: Plua_State; narg: Integer); cdecl; external LUA_LIB_NAME;

procedure luaL_where(L: Plua_State; level: Integer); cdecl; external LUA_LIB_NAME;
function luaL_error(L: Plua_State; const fmt: PChar; argsup: array of pointer): Integer; cdecl; external LUA_LIB_NAME;

function luaL_findstring(name: PChar; const list: PPChar): Integer; cdecl; external LUA_LIB_NAME;

function luaL_ref(L: Plua_State; t: Integer): Integer; cdecl; external LUA_LIB_NAME;
procedure luaL_unref(L: Plua_State; t, ref: Integer); cdecl; external LUA_LIB_NAME;

function luaL_loadfile(L: Plua_State; const filename: PChar): Integer; cdecl; external LUA_LIB_NAME;
function luaL_loadbuffer(L: Plua_State; const buff: PChar; size: Integer; const name: PChar): Integer; cdecl; external LUA_LIB_NAME;

procedure luaL_arg_check(L: Plua_State; cond: Boolean; numarg: Integer; extramsg: PChar);
begin
  if not cond then
    luaL_argerror(L, numarg, extramsg)
end;

function luaL_check_string(L: Plua_State; n: Integer): PChar;
begin
  Result := luaL_check_lstr(L, n, nil)
end;

function luaL_opt_string(L: Plua_State; n: Integer; d: PChar): PChar;
begin
  Result := luaL_opt_lstr(L, n, d, nil)
end;

function luaL_check_int(L: Plua_State; n: Integer): Integer;
begin
  Result := Integer(Trunc(luaL_check_number(L, n)))
end;

function luaL_check_long(L: Plua_State; n: Integer): LongInt;
begin
  Result := LongInt(Trunc(luaL_check_number(L, n)))
end;

function luaL_opt_int(L: Plua_State; n: Integer; d: Double): Integer;
begin
  Result := Integer(Trunc(luaL_opt_number(L, n, d)))
end;

function luaL_opt_long(L: Plua_State; n: Integer; d: Double): LongInt;
begin
  Result := LongInt(Trunc(luaL_opt_number(L, n, d)))
end;

procedure luaL_putchar(B: PluaL_Buffer; c: Char);
begin
  if Cardinal(@(B^.p)) < (Cardinal(@(B^.buffer[0])) + LUAL_BUFFERSIZE) then
    luaL_prepbuffer(B);
  B^.p[1] := c;
  B^.p := B^.p + 1;
end;

procedure luaL_addsize(B: PluaL_Buffer; n: Integer);
begin
  B^.p := B^.p + n;
end;

procedure luaL_buffinit(L: Plua_State ; B: PluaL_Buffer); cdecl; external LUA_LIB_NAME;
function luaL_prepbuffer(B: PluaL_Buffer): PChar; cdecl; external LUA_LIB_NAME;
procedure luaL_addlstring(B: PluaL_Buffer; const s: PChar; l: size_t); cdecl; external LUA_LIB_NAME;
procedure luaL_addstring(B: PluaL_Buffer; const s: PChar); cdecl; external LUA_LIB_NAME;
procedure luaL_addvalue(B: PluaL_Buffer); cdecl; external LUA_LIB_NAME;
procedure luaL_pushresult(B: PluaL_Buffer); cdecl; external LUA_LIB_NAME;

procedure luaL_checktype(L: Plua_State; narg, t: Integer);
begin
  luaL_checktype(L, narg, t);
end;

procedure luaL_checkany(L: Plua_State; narg: Integer);
begin
  luaL_check_any(L, narg);
end;

function lua_dofile(L: Plua_State; const filename: PChar): Integer; cdecl; external LUA_LIB_NAME;
function lua_dostring(L: Plua_State; const str: PChar): Integer; cdecl; external LUA_LIB_NAME;
function lua_dobuffer(L: Plua_State; const buff: PChar; size: Integer; const name: PChar): Integer; cdecl; external LUA_LIB_NAME;

end.