unit unitdesktopview;

(* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1
 *
 * The contents of this file are subject to the Mozilla Public License Version 
 * 1.1 (the "License"); you may not use this file except in compliance with 
 * the License. You may obtain a copy of the License at 
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is the Lightning TDesktopView class.
 *
 * The Initial Developer of the Original Code is
 * Tomsoft.
 * Portions created by Tomsoft are Copyright (C) 2009
 * Tomsoft. All Rights Reserved.
 *
 * Contributor(s):
 *   Tommy Sools <tom@sools.com>
 *
 * ***** END LICENSE BLOCK ***** *)

interface

uses
  Forms, Windows, Messages, Classes, Controls, Graphics, ExtCtrls, Xdom_3_2,
  common, unitmodule, unitconnector;

type
  TDesktopView = class(TScrollBox)
  private
    paintDisabled:   Integer;
    focusRect:       TRect;
    selectionStart:  TPoint;
    procedure desktopViewDragOver(Sender, Source: TObject; X,
      Y: Integer; State: TDragState; var Accept: Boolean);
    procedure desktopViewDragDrop(Sender, Source: TObject; X, Y: Integer);
    procedure desktopViewMouseDown(Sender: TObject;
      Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
    procedure desktopViewMouseMove(Sender: TObject;
      Shift: TShiftState; X, Y: Integer);
    procedure desktopViewMouseUp(Sender: TObject;
      Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
  protected
    procedure WMPAINT(var Message: TMessage); message WM_PAINT;
  public
    // array containing all modules of desktop
    modules: array of TModule;
    // array containing only the selected modules of desktop
    selectedModules: array of TModule;
    highlightOutput: TOutput;
    deltaX, deltaY: Integer;
    // popup variable stores position of context menu,
    // so that the position can be used for a newly created module.
    popup: TPoint;
    contextModule: TModule;
    constructor Create(AOwner: TComponent); override;
    procedure updateConnectors();
    procedure updateLayers();
    procedure updateLocked();
    procedure enablePaint();
    procedure disablePaint();
    procedure paintLines();
    procedure makeFocusRect(r: TRect);
    procedure clearFocusRect();
    procedure mouseOut();
    function generateSceneid(): string;
    function findModule(id: string; mapModuleId: TStringList): TModule;
    function findModuleName(name: string): TModule;
    function findModuleIndex(module: TModule): Integer;
    procedure clear();
    procedure addFromXML(xmlNode: TDOMNode; offsetTop, offsetLeft: Integer; isClipboard: Boolean; isSelected: Boolean; visible: Boolean);
    procedure addFromXMLFile(fileName, rootElement: string);
    procedure addFromXMLString(s, rootElement: string; offsetTop, offsetLeft: Integer; isClipboard: Boolean; isSelected: Boolean; visible: Boolean);
    function toXML(rootElement: string; selectedOnly: Boolean): string;
    procedure toXMLFile(fileName, rootElement: string);
    procedure fromStringFile(fileName: string);
    function toString(selectedOnly, includeModuleId: Boolean): string;
    procedure toStringFile(fileName: string);
    procedure moduleAction();
    procedure allModulesSetIsSelected(isSelected: Boolean);
    procedure allModulesKeyboard(nVirtKey: Longint; lKeyData: Cardinal);
    function selectedModulesGetBounds(var top, left, width, height: Integer): Boolean;
    procedure selectedModulesCut();
    procedure selectedModulesCopy();
    procedure paste(xCorrection, yCorrection: Integer);
    procedure selectedModulesDelete();
    procedure selectedModulesSetVisible(visible: Boolean);
    procedure selectedModulesSetLayer(layer: Integer);
    procedure selectedModulesBringToFront();
    procedure selectedModulesSendToBack();
    function selectedModulesGetColor(): TColor;
    procedure selectedModulesSetColor(color: TColor);
    procedure selectedModulesSetIsUsedForScene(usedForScene: Boolean);
  end;

procedure initClipboard();

implementation

uses
  SysUtils, unitlightningsettings, unitformlightning, unitformtrace;

var
  CF_LIGHTNING_MODULE, CF_LIGHTNING_MODULE_XML: Word;

procedure initClipboard();
begin
  CF_LIGHTNING_MODULE := RegisterClipboardFormat('Lightning Module');
  CF_LIGHTNING_MODULE_XML := RegisterClipboardFormat('XML Lightning Module');
end;

procedure copyToClipboard(uFormat: Cardinal; content: string);
{
  Pre:  OpenClipboard was successfully called;
  Post: content was copied to the clipboard with format uFormat;
}
var
  data:     THandle;
  dataPtr:  Pointer;
  size:     Integer;
begin
  size := Length(content) + 1;
  data := GlobalAlloc(GMEM_MOVEABLE or GMEM_DDESHARE, size);
  try
    dataPtr := GlobalLock(data);
    try
      Move(PChar(content)^, dataPtr^, size);
      SetClipboardData(uFormat, data);
    finally
      GlobalUnlock(data);
    end;
  except
    GlobalFree(data);
    raise;
  end;
end;

procedure pasteFromClipboard(uFormat: Cardinal; var content: string);
{
  Pre:  OpenClipboard was successfully called;
  Post: the content of clipboard format uFormat was copied to content;
}
var
  data: THandle;
  pData: PChar;
begin
  content := '';
  data := GetClipboardData(uFormat);
  if data <> 0 then
  begin
    pData := PChar(GlobalLock(data));
    if pData <> nil then
    begin
      content := Trim(pData);
      GlobalUnlock(data);
    end;
  end;
end;

procedure reorderRect(var r: TRect);
var
  bInteger: Integer;
begin
  // reorder coordinates for rectangle
  if (r.Left > r.Right) then
  begin
    bInteger := r.Left;
    r.Left   := r.Right;
    r.Right  := bInteger;
  end;
  if (r.Top > r.Bottom) then
  begin
    bInteger := r.Top;
    r.Top    := r.Bottom;
    r.Bottom := bInteger;
  end;
end;

function insideRect(control: TControl; r: TRect): Boolean;
begin
  Result :=
    (control.Top + control.Height > r.Top)
    and
    (control.Top < r.Bottom)
    and
    (control.Left + control.Width > r.Left)
    and
    (control.Left < r.Right);
end;

constructor TDesktopView.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  BevelInner      := bvNone;
  BevelOuter      := bvNone;
  BorderStyle     := bsNone;
  Align           := alClient;
  Color           := clBtnFace;
  paintDisabled   := 0;
  selectionStart  := Point(MaxInt, MaxInt);
  focusRect       := Rect(MaxInt, MaxInt, MaxInt, MaxInt);
  highlightOutput := nil;
  contextModule   := nil;
  OnDragOver      := desktopViewDragOver;
  OnDragDrop      := desktopViewDragDrop;
  OnMouseDown     := desktopViewMouseDown;
  OnMouseMove     := desktopViewMouseMove;
  OnMouseUp       := desktopViewMouseUp;
end;

procedure TDesktopView.WMPAINT(var Message: TMessage);
begin
  inherited;
  paintLines();
end;

// update the color and label of all input and output connectors
procedure TDesktopView.updateConnectors();
var
  i, j: Integer;
begin
  for i := 0 to Length(modules) - 1 do
  begin
    for j := 1 to modules[i].module^.inputs do
      modules[i].input[j].clearLevel();
    for j := 1 to modules[i].module^.outputs do
      modules[i].output[j].clearLevel();
  end;
end;

procedure TDesktopView.updateLayers();
var
  i: Integer;
begin
  for i := 0 to Length(modules) - 1 do
    modules[i].updateLayer();
  Repaint();
end;

procedure TDesktopView.updateLocked();
var
  i: Integer;
begin
  for i := 0 to Length(modules) - 1 do
    modules[i].setLocked(lightningSettings.lock);
end;

procedure TDesktopView.enablePaint();
begin
  if paintDisabled > 0 then
    Dec(paintDisabled);
  paintLines();
end;

procedure TDesktopView.disablePaint();
begin
  Inc(paintDisabled);
end;

procedure TDesktopView.paintLines();
var
  i, j: Integer;
begin
  if paintDisabled < 1 then
  begin
    if lightningSettings.paintUnpoweredLines then
      for i := 0 to Length(modules) - 1 do
        for j := 1 to modules[i].module^.outputs do
          modules[i].output[j].paintLine(clGray, R2_COPYPEN, False);
  end;
end;

// this visualizes the focus rectangle for module dragging
procedure TDesktopView.desktopViewDragOver(Sender, Source: TObject; X,
  Y: Integer; State: TDragState; var Accept: Boolean);
var
  bPoint: TPoint;
  movementX, movementY, selTop, selLeft, selWidth, selHeight: Integer;
begin
  if Source is TModule then
  begin
    bPoint.Y := Y;
    bPoint.X := X;
    if Sender <> Self then
      bPoint := TControl(Sender).ClientToParent(bPoint, Self);
    movementY := bPoint.Y - {TDesktopView(TModule(Source).Parent).}deltaY - TModule(Source).Top;
    movementX := bPoint.X - {TDesktopView(TModule(Source).Parent).}deltaX - TModule(Source).Left;
    if not(selectedModulesGetBounds(selTop, selLeft, selWidth, selHeight)) then
    begin
      selTop := TModule(Source).Top;
      selLeft := TModule(Source).Left;
      selWidth := TModule(Source).Width;
      selHeight := TModule(Source).Height;
    end;
    // replace old focus rectangle with a new one
    if State <> dsDragEnter then
      clearFocusRect();
    if State <> dsDragLeave then
      makeFocusRect(
        Rect(
          selLeft + movementX,
          selTop + movementY,
          selLeft + selWidth + movementX,
          selTop + selHeight + movementY
        )
      );
    Accept := True;
  end
  else
    Accept := False;
end;

procedure TDesktopView.desktopViewDragDrop(Sender, Source: TObject; X, Y: Integer);
var
  bPoint: TPoint;
  movementX, movementY, i: Integer;
  desktopView: TDesktopView;
begin
  if Source is TModule then
  begin
    desktopView := TDesktopView(TModule(Source).Parent);
    bPoint.Y := Y;
    bPoint.X := X;
    if Sender <> Self then
      bPoint := TControl(Sender).ClientToParent(bPoint, Self);
    // calculate the dragging offset
    movementY := bPoint.Y - desktopView.deltaY - TControl(Source).Top;
    movementX := bPoint.X - desktopView.deltaX - TControl(Source).Left;
    // reposition the dragged control
    with TModule(Source) do
      SetBounds(Left + movementX, Top + movementY, Width, Height);
    // reposition any other selected controls
    for i := 0 to Length(desktopView.selectedModules) - 1 do
      if desktopView.selectedModules[i] <> TModule(Source) then
        with desktopView.selectedModules[i] do
          SetBounds(Left + movementX, Top + movementY, Width, Height);
  end;
end;

// this deselects all modules and sets the user selection start position
procedure TDesktopView.desktopViewMouseDown(Sender: TObject;
  Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
  if not(lightningSettings.lock) then
  begin
    case Button of
      mbLeft:
      begin
        if not(ssCtrl in Shift) then
          allModulesSetIsSelected(False);
        selectionStart.X := X;
        selectionStart.Y := Y;
      end;
    end;
  end;
end;

// this updates the focus rectangle for user selection
procedure TDesktopView.desktopViewMouseMove(Sender: TObject;
  Shift: TShiftState; X, Y: Integer);
var
  r: TRect;
begin
  mouseOut();
  if selectionStart.X < MaxInt then // check if the user started dragging at all
  begin
    clearFocusRect();
    // store coordinates for user selection rectangle and make sure they are ordered appropriately
    r := Rect(selectionStart, Point(X, Y));
    reorderRect(r);
    makeFocusRect(r);
  end;
end;

procedure TDesktopView.desktopViewMouseUp(Sender: TObject;
  Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
var
  selectionRect: TRect;
  i:             Integer;
begin
  if not(lightningSettings.lock) then
  begin
    case Button of
      mbLeft:
      begin
        // adjust module selection
        if selectionStart.X < MaxInt then // check if the user started dragging at all
        begin
          clearFocusRect();
          // store coordinates of selection and make sure they are ordered appropriately
          selectionRect := Rect(selectionStart, Point(X, Y));
          reorderRect(selectionRect);
          // clear the user selection start position
          selectionStart.X := MaxInt;
          selectionStart.Y := MaxInt;
          // select all modules that are inside (or just even "touch") the selection rectangle
          for i := 0 to Length(modules) - 1 do
            if insideRect(modules[i], selectionRect) then
              modules[i].setIsSelected(True);
        end;
      end;
      mbRight:
      begin
        // show context menu at cursor position
        GetCursorPos(popup);
        formLightning.pmPopup.Popup(popup.X, popup.Y);
        popup := ScreenToClient(popup);
      end;
    end;
  end;
end;

procedure TDesktopView.makeFocusRect(r: TRect);
{
  Pre:  no focus rectangle exists;
  Post: a focus rectangle r exists;
}
var
  bDC: hDC;
begin
  focusRect := r;
  bDC := GetDC(Handle);
  try
    DrawFocusRect(bDC, r);
  finally
    ReleaseDC(Handle, bDC);
  end;
end;

procedure TDesktopView.clearFocusRect();
{
  Pre:  true;
  Post: no focus rectangle exists;
}
var
  bDC: hDC;
begin
  if focusRect.Left < MaxInt then // check if an old focus rectangle exists
  begin
    // an old focus rectangle exists, so remove it
    bDC := GetDC(Handle);
    try
      DrawFocusRect(bDC, focusRect);
    finally
      ReleaseDC(Handle, bDC);
    end;
    focusRect := Rect(MaxInt, MaxInt, MaxInt, MaxInt);
  end;
end;

procedure TDesktopView.mouseOut();
begin
  if highlightOutput <> nil then
    highlightOutput.unHighlight();
  // restore status bar text
  formLightning.statusBar.Panels[2].Text := '';
  formLightning.statusBar.Panels[3].Text := '';
  formLightning.statusBar.Panels[4].Text := '';
end;

function TDesktopView.generateSceneid(): string;
{
  Pre:  true;
  Post: Result = a sceneid not used by any module in the desktopview yet;
}
var
  i:      Integer;
  unique: Boolean;
begin
  repeat
    Result := IntToStr(Random(99999999));
    unique := True;
    for i := 0 to Length(modules) - 1 do
      if modules[i].sceneid = Result then
      begin
        unique := False;
        Break;
      end;
  until unique;
end;

function TDesktopView.findModule(id: string; mapModuleId: TStringList): TModule;
var
  i, mapIndex: Integer;
begin
  Result := nil;
  if mapModuleId <> nil then
  begin
    // if a valid translation for the module id exists,
    // due to pasting from the clipboard, use it.
    mapIndex := mapModuleId.IndexOfObject(TObject(StrToInt(id)));
    if mapIndex > -1 then
      id := mapModuleId.Strings[mapIndex];
  end;
  for i := 0 to Length(modules) - 1 do
    if modules[i].sceneid = id then
    begin
      Result := modules[i];
      Break;
    end;
end;

function TDesktopView.findModuleName(name: string): TModule;
var
  i: Integer;
begin
  Result := nil;
  for i := 0 to Length(modules) - 1 do
    if modules[i].ext = name then
    begin
      Result := modules[i];
      Break;
    end;
end;

function TDesktopView.findModuleIndex(module: TModule): Integer;
var
  i: Integer;
begin
  Result := -1;
  for i := 0 to Length(modules) - 1 do
    if modules[i] = module then
    begin
      Result := i;
      Break;
    end;
end;

procedure TDesktopView.clear();
var
  module: TModule;
begin
  Visible := False;
  try
    // remove all modules
    while Length(modules) > 0 do
    begin
      module := modules[0];
      module.Free();
    end;
  finally
    Visible := True;
  end;
end;

procedure TDesktopView.addFromXML(xmlNode: TDOMNode; offsetTop, offsetLeft: Integer; isClipboard: Boolean; isSelected: Boolean; visible: Boolean);
var
  xmlAttrNode, xmlChildNode: TDOMNode;
  version, id: string;
  module: TModule;
  mapModuleId: TStringList;
begin
  // read attributes
  xmlAttrNode := xmlNode.Attributes.GetNamedItem('version');
  if xmlAttrNode <> nil then
  begin
    version := xmlAttrNode.TextContent;
    // TODO Raise error if file is too new
  end;
  if isClipboard then
    mapModuleId := TStringList.Create()
  else
    mapModuleId := nil;
  // read modules
  xmlChildNode := xmlNode.GetFirstChildElement('modules');
  if xmlChildNode <> nil then
  begin
    // walk thru all elements inside the XML element
    xmlChildNode := xmlChildNode.FindFirstChildElement();
    while xmlChildNode <> nil do
    begin
      if xmlChildNode.NodeName = 'module' then
        TModule.createFromXML(xmlChildNode, Self, offsetTop, offsetLeft, mapModuleId, isSelected, visible);
      xmlChildNode := xmlChildNode.FindNextSiblingElement();
    end;
  end;
  // read connections for each module
  xmlChildNode := xmlNode.GetFirstChildElement('modules');
  if xmlChildNode <> nil then
  begin
    // walk thru all elements inside the XML element
    xmlChildNode := xmlChildNode.FindFirstChildElement();
    while xmlChildNode <> nil do
    begin
      if xmlChildNode.NodeName = 'module' then
      begin
        // read attributes
        xmlAttrNode := xmlChildNode.Attributes.GetNamedItem('id');
        if xmlAttrNode <> nil then
          id := xmlAttrNode.TextContent;
        if id <> '' then
        begin
          module := findModule(id, mapModuleId);
          if module <> nil then
            module.addConnectionsFromXML(xmlChildNode, mapModuleId)
          else
            raise Exception.Create('There was an error reading the XML content. The module referenced to by an output could not be found.');
        end;
      end;
      xmlChildNode := xmlChildNode.FindNextSiblingElement();
    end;
  end;
  if mapModuleId <> nil then
    mapModuleId.Free();
end;

procedure TDesktopView.addFromXMLFile(fileName, rootElement: string);
var
  xmlParser: TXmlToDomParser;
  xmlDocument: TDOMDocument;
  xmlNode: TDOMNode;
begin
  Visible := False;
  try
    try
      xmlParser := TXmlToDomParser.Create(nil);
      try
        xmlParser.DOMImpl := TDomImplementation.Create(nil);
        try
          xmlDocument := xmlParser.FileToDom(fileName);
          try
            xmlNode := xmlDocument.FindFirstChildElement();
            if xmlNode <> nil then
            begin
              if xmlNode.NodeName = rootElement then
              begin
                addFromXML(xmlNode, 0, 0, False, False, False);
              end;
            end;
          finally
            xmlDocument.Free();
          end;
        finally
          xmlParser.DOMImpl.Free();
          xmlParser.DOMImpl := nil;
        end;
      finally
        xmlParser.Free();
      end;
    except
      on E: EParserException do
        raise Exception.Create('There was an error reading the XML file. Make sure the XML file has the right layout.');
    end;
  finally
    Visible := True;
  end;
end;

procedure TDesktopView.addFromXMLString(s, rootElement: string; offsetTop, offsetLeft: Integer; isClipboard: Boolean; isSelected: Boolean; visible: Boolean);
var
  xmlParser: TXmlToDomParser;
  xmlDocument: TDOMDocument;
  xmlNode: TDOMNode;
begin
  Self.Visible := False;
  try
    xmlParser := TXmlToDomParser.Create(nil);
    try
      xmlParser.DOMImpl := TDomImplementation.Create(nil);
      try
        xmlDocument := xmlParser.StringToDom(s, '', nil, True);
        try
          xmlNode := xmlDocument.FindFirstChildElement();
          if xmlNode <> nil then
          begin
            if xmlNode.NodeName = rootElement then
            begin
              addFromXML(xmlNode, offsetTop, offsetLeft, isClipboard, isSelected, visible);
            end;
          end;
        finally
          xmlDocument.Free();
        end;
      finally
        xmlParser.DOMImpl.Free();
        xmlParser.DOMImpl := nil;
      end;
    finally
      xmlParser.Free();
    end;
  except
    on E: EParserException do
      raise Exception.Create('There was an error reading the XML content. Make sure the XML content has the right layout.');
  end;
  Self.Visible := True;
end;

function TDesktopView.toXML(rootElement: string; selectedOnly: Boolean): string;
var
  i: Integer;
begin
  Result :=
    '<?xml version="1.0" encoding="UTF-8"?>'#13#10 +
    '<' + rootElement + ' version="1.0">'#13#10 +
    '<modules>'#13#10;
  if selectedOnly then
  begin
    for i := 0 to Length(selectedModules) - 1 do
      Result := Result + selectedModules[i].toXML() + #13#10;
  end
  else
  begin
    for i := 0 to Length(modules) - 1 do
      Result := Result + modules[i].toXML() + #13#10;
  end;
  Result :=
    Result +
    '</modules>'#13#10 +
    '</' + rootElement + '>'#13#10;
end;

procedure TDesktopView.toXMLFile(fileName, rootElement: string);
var
  output: TStringList;
begin
  output := TStringList.Create();
  try
    output.Text := toXML(rootElement, False);
    output.SaveToFile(fileName);
  finally
    output.Free();
  end;
end;

procedure TDesktopView.fromStringFile(fileName: string);
var
  input: TStringList;
  line:  string;
  error: Boolean;
  m:     TModule;
  i, c, j, l, connectorCount, module, inputno, iError, iError2: Integer;
begin
  Visible := False;
  try
    input := TStringList.Create();
    try
      input.LoadFromFile(filename);
      if input.Count > 0 then
        if input.Count > 1 then
          if input[0] = 'Lightning|1.0' then
          begin
            c := StrToIntDef(input[1], 0);
            if c > 0 then
            begin
              // load modules
              error := False;
              i     := 0;
              while not(error) and (i < c) do
              begin
                if input.Count > 2 + i then
                begin
                  m := TModule.createFromString(input[2 + i], Self, 0, 0, False, True);
                  error := m = nil;
                  Inc(i);
                end
                else
                begin
                  MessageBox(Handle,
                    'Corrupt file: missing module specification!',
                    'Error',
                     MB_ICONERROR
                  );
                  error := True;
                end;
              end;
              if not(error) then
              begin
                // load module connections
                i := 0;
                l := 2 + c;
                while not(error) and (i < c) do
                begin
                  if input.Count > l then
                  begin
                    connectorCount := StrToIntDef(input[l], 0);
                    j := 1;
                    while not(error) and (j <= connectorCount) do
                    begin
                      Inc(l);
                      if input.Count > l then
                      begin
                        line := input[l];
                        if line <> '' then
                        begin
                          Val(parse(line, '|'), module,  iError);
                          Val(parse(line, '|'), inputno, iError2);
                          if (iError = 0) and (iError2 = 0) then
                          begin
                            if j <= modules[i].module^.outputs then
                            begin
                              if inputno <= modules[module].module^.inputs then
                              begin
                                modules[i].output[j].connect(modules[module], inputno);
                              end;
                              // else the saved file contains a reference to a module's input link that doesn't exist in this version of that module.
                              // the problem will be ignored to improve backwards compatibility, for example related to the OUT_DMX module.
                            end
                            else
                            begin
                              // the saved file contains more connections than the module can handle
                              MessageBox(Handle,
                                'Corrupt file: the module contains specifications for more output links than exist in the module!',
                                'Error',
                                MB_ICONERROR
                              );
                              error := True;
                            end;
                          end
                          else
                          begin
                            MessageBox(Handle,
                              'Corrupt file: invalid module output link specification!',
                              'Error',
                              MB_ICONERROR
                            );
                            error := True;
                          end;
                        end;
                        Inc(j);
                      end
                      else
                      begin
                        MessageBox(Handle,
                          'Corrupt file: missing module output link specification!',
                          'Error',
                          MB_ICONERROR
                        );
                        error := True;
                      end;
                    end;
                    Inc(i);
                    Inc(l);
                  end
                  else
                  begin
                    MessageBox(Handle,
                      'Corrupt file: missing module link specification!',
                      'Error',
                       MB_ICONERROR
                    );
                    error := True;
                  end;
                end;
              end;
            end;
          end
          else
            MessageBox(Handle,
              'Corrupt file: invalid file format!',
              'Error',
              MB_ICONERROR
            )
        else
          MessageBox(Handle,
            'Corrupt file: invalid line count!',
            'Error',
            MB_ICONERROR
          )
      else
        MessageBox(Handle,
          'Corrupt file: file is empty!',
          'Error',
          MB_ICONERROR
        );
    finally
      input.Free();
    end;
  finally
    Visible := True;
  end;
end;

function TDesktopView.toString(selectedOnly, includeModuleId: Boolean): string;
var
  i: Integer;
begin
  Result := '';
  if selectedOnly then
  begin
    for i := 0 to Length(selectedModules) - 1 do
      Result := Result + selectedModules[i].toString(includeModuleId) + #13#10;
  end
  else
  begin
    for i := 0 to Length(modules) - 1 do
      Result := Result + modules[i].toString(includeModuleId) + #13#10;
  end;
end;

procedure TDesktopView.toStringFile(fileName: string);
var
  output: TStringList;
  i, j, k: Integer;
  line: string;
  bOutput: TOutput;
begin
  output := TStringList.Create();
  try
    // add header
    output.Add('Lightning|1.0');
    // add modules
    output.Add(IntToStr(Length(modules)));
    output.Add(TrimRight(toString(False, True)));
    // add module connections
    for i := 0 to Length(modules) - 1 do
    begin
      output.Add(IntToStr(modules[i].module^.outputs));
      for j := 1 to modules[i].module^.outputs do
      begin
        line := '';
        bOutput := modules[i].output[j];
        if bOutput.isConnected() then
        begin
          k := findModuleIndex(bOutput.nextModule);
          if k > -1 then
            line := IntToStr(k) + '|' + IntToStr(bOutput.nextInput)
          else
            MessageBox(Handle, 'There was a problem with saving the file,'#13#10'some information may be lost!', 'Warning', MB_ICONERROR);
        end;
        output.Add(line);
      end;
    end;
    output.SaveToFile(fileName);
  finally
    output.Free();
  end;
end;

procedure TDesktopView.moduleAction();
var
  i: Integer;
  dummy: TSpectrum;
begin
  if formTrace <> nil then
  begin
    formTrace.start();
    formTrace.add('moduleAction(start)');
  end;
  if Length(modules) > 0 then
    for i := 0 to Length(modules) - 1 do
    begin
      case modules[i].kind of
        mkComponent:
        begin
          // process all Input modules inside a component
          TDesktopView(modules[i].module^.details).moduleAction();
        end;
        mkComponentIn:
        begin
          // only let the in_component module act if it is on the main desktop,
          // so that it can output zero values
          if modules[i].module^.details = nil then
          begin
            createSpectrum(dummy, levelMin);
            modules[i].actionInput(0, dummy, True);
          end;
        end;
      else
        if ((modules[i].module^.inputs = 0) and (modules[i].module^.outputs > 0)) then
        begin
          // only let the module act on the timer if it is on the root of the chain,
          // meaning it has no inputs while it does have any outputs
          modules[i].actionInput(0, dummy, True)
        end
      end;
      // in the future also act on module inputs that are unconnected
    end;
  if formTrace <> nil then
  begin
    formTrace.add('moduleAction(end)');
    formTrace.stop();
  end;
end;

procedure TDesktopView.allModulesSetIsSelected(isSelected: Boolean);
{
  Pre:  true;
  Post: if isSelected ==> (all visible modules on the current desktop are selected);
        if not(isSelected) ==> (all modules on the current desktop are unselected);
}
var
  i: Integer;
begin
  if isSelected then
  begin
    for i := 0 to Length(modules) - 1 do
      modules[i].setIsSelected(modules[i].Visible);
  end
  else
  begin
    while Length(selectedModules) > 0 do
      selectedModules[0].setIsSelected(False);
  end;
end;

procedure TDesktopView.allModulesKeyboard(nVirtKey: Longint; lKeyData: Cardinal);
var
  i: Integer;
begin
  for i := 0 to Length(modules) - 1 do
    if @modules[i].keyboard <> nil then
      modules[i].keyboard(modules[i].module, nVirtKey, lKeyData);
end;

function TDesktopView.selectedModulesGetBounds(var top, left, width, height: Integer): Boolean;
var
  minTop, minLeft, maxRight, maxBottom, i: Integer;
begin
  minTop := MaxInt;
  minLeft := MaxInt;
  maxRight := -MaxInt - 1;
  maxBottom := -MaxInt - 1;;
  Result := False;
  for i := 0 to Length(selectedModules) - 1 do
  begin
    Result := True;
    if selectedModules[i].Top < minTop then
      minTop := selectedModules[i].Top;
    if selectedModules[i].Left < minLeft then
      minLeft := selectedModules[i].Left;
    if (selectedModules[i].Left + selectedModules[i].Width) > maxRight then
      maxRight := selectedModules[i].Left + selectedModules[i].Width;
    if (selectedModules[i].Top + selectedModules[i].Height) > maxBottom then
      maxBottom := selectedModules[i].Top + selectedModules[i].Height;
  end;
  if minLeft = MaxInt then
  begin
    left := 0;
    width := 0;
  end
  else
  begin
    left := minLeft;
    width := maxRight - minLeft;
  end;
  if minTop = MaxInt then
  begin
    top := 0;
    height := 0;
  end
  else
  begin
    top := minTop;
    height := maxBottom - minTop;
  end;
end;

procedure TDesktopView.selectedModulesCut();
begin
  selectedModulesCopy();
  selectedModulesDelete();
end;

procedure TDesktopView.selectedModulesCopy();
var
  i: Integer;
  bufferXML: string;
begin
  // copy all selected modules to the clipboard
  i := 0;
  while not(OpenClipboard(Handle)) and (i < 100) do
  begin
    Sleep(100);
    Inc(i);
  end;
  if i < 100 then
  begin
    try
      EmptyClipboard();
      bufferXML := toXML('lightningclipboard', True);
      copyToClipboard(CF_TEXT, bufferXML);
      copyToClipboard(CF_LIGHTNING_MODULE, toString(True, False));
      copyToClipboard(CF_LIGHTNING_MODULE_XML, bufferXML);
    finally
      CloseClipboard();
    end;
  end;
end;

procedure TDesktopView.paste(xCorrection, yCorrection: Integer);
var
  j: Integer;
  bStringList: TStringList;
  buffer: string;
begin
  j := 0;
  // try to open the clipboard for a maximum of 10 seconds
  while not(OpenClipboard(Handle)) and (j < 100) do
  begin
    Sleep(100);
    Inc(j);
  end;
  if j < 100 then
  begin
    try
      pasteFromClipboard(CF_LIGHTNING_MODULE_XML, buffer);
      if buffer <> '' then
        // for XML clipboard format
        addFromXMLString(buffer, 'lightningclipboard', yCorrection, xCorrection, True, True, False)
      else
      begin
        // for clipboard format backward compatibility
        bStringList := TStringList.Create();
        try
          pasteFromClipboard(CF_LIGHTNING_MODULE, buffer);
          bStringList.Text := buffer;
          for j := 0 to bStringList.Count - 1 do
            TModule.createFromString(bStringList[j], Self, yCorrection, xCorrection, True, False);
        finally
          bStringList.Free();
        end;
      end;
    finally
      CloseClipboard();
    end;
  end;
end;

procedure TDesktopView.selectedModulesDelete();
begin
  while Length(selectedModules) > 0 do
    formLightning.moduleDelete(selectedModules[0]);
end;

procedure TDesktopView.selectedModulesSetVisible(visible: Boolean);
var
  i: Integer;
begin
  for i := 0 to Length(selectedModules) - 1 do
    selectedModules[i].Visible := visible;
end;

procedure TDesktopView.selectedModulesSetLayer(layer: Integer);
var
  i: Integer;
  module: TModule;
begin
  i := 0;
  while i < Length(selectedModules) do
  begin
    module := selectedModules[i];
    module.setLayer(layer);
    // invisible modules became unselected after calling setLayer
    if module.Visible then
      Inc(i);
  end;
  Repaint();
end;

procedure TDesktopView.selectedModulesBringToFront();
var
  i: Integer;
begin
  for i := 0 to Length(selectedModules) - 1 do
    selectedModules[i].BringToFront();
end;

procedure TDesktopView.selectedModulesSendToBack();
var
  i: Integer;
begin
  for i := 0 to Length(selectedModules) - 1 do
    selectedModules[i].SendToBack();
end;

function TDesktopView.selectedModulesGetColor(): TColor;
var
  i: Integer;
begin
  Result := clBtnFace;
  for i := 0 to Length(selectedModules) - 1 do
  begin
    Result := selectedModules[i].getColor();
    if Result <> clBtnFace then
      Break;
  end;
end;

procedure TDesktopView.selectedModulesSetColor(color: TColor);
var
  i: Integer;
begin
  for i := 0 to Length(selectedModules) - 1 do
    selectedModules[i].setColor(color);
end;

procedure TDesktopView.selectedModulesSetIsUsedForScene(usedForScene: Boolean);
var
  i: Integer;
begin
  for i := 0 to Length(selectedModules) - 1 do
    selectedModules[i].setIsUsedForScene(usedForScene);
end;

end.
