unit unitformprogrammanager;

(* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1
 *
 * The contents of this file are subject to the Mozilla Public License Version 
 * 1.1 (the "License"); you may not use this file except in compliance with 
 * the License. You may obtain a copy of the License at 
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is the Lightning program manager window.
 *
 * The Initial Developer of the Original Code is
 * Tomsoft.
 * Portions created by Tomsoft are Copyright (C) 2009
 * Tomsoft. All Rights Reserved.
 *
 * Contributor(s):
 *   Tommy Sools <tom@sools.com>
 *
 * ***** END LICENSE BLOCK ***** *)

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, StdCtrls, Menus, ToolWin, ComCtrls, ImgList,
  common, unitformlightning, unitformscenemanager;

type
  TformProgramManager = class(TForm)
    lbPrograms: TListBox;
    lFadeTime: TLabel;
    lFadeTimeNotice: TLabel;
    chOverride: TCheckBox;
    chNoProgramSwitch: TCheckBox;
    pmFadeTime: TPopupMenu;
    miApplyToAllPrograms: TMenuItem;
    tbProgramManager: TToolBar;
    tbNew: TToolButton;
    tbOpen: TToolButton;
    tbRename: TToolButton;
    tbDuplicate: TToolButton;
    tbDelete: TToolButton;
    tbSep1: TToolButton;
    tbUp: TToolButton;
    tbDown: TToolButton;
    tbSep2: TToolButton;
    upDownFadeTime: TUpDown;
    editFadeTime: TEdit;
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
    procedure miApplyToAllProgramsClick(Sender: TObject);
    procedure tbNewClick(Sender: TObject);
    procedure tbOpenClick(Sender: TObject);
    procedure tbRenameClick(Sender: TObject);
    procedure tbDuplicateClick(Sender: TObject);
    procedure tbDeleteClick(Sender: TObject);
    procedure tbUpClick(Sender: TObject);
    procedure tbDownClick(Sender: TObject);
  private
    { Private declarations }
    //function  saveProgram(noConfirm: Boolean): Boolean;
    function  generateFileExt(i: Integer): string;
  public
    { Public declarations }
    function  numberToFileName(base: string; number: Integer): string;
    function  indexToFileName(base: string; index: Integer): string;
    function  determineFreeProgramNumber(): Integer;
    function  requireDesktopSaved(noConfirm: Boolean): Boolean;
    function  saveCurrentProgram(noConfirm: Boolean): Boolean;
    function  newProgram(noConfirm: Boolean): Boolean;
    function  switchProgram(index: Integer; noConfirm: Boolean): Boolean;
    procedure loadProgram(index: Integer);
    function  getProgramName(index: Integer): string;
  end;

var
  formProgramManager:  TformProgramManager;
  currentProgramIndex: Integer = -1;
  currentProgramName:  string = '';

implementation

{$R *.dfm}

function  TformProgramManager.generateFileExt(i: Integer): string;
begin
  Result := IntToStr(i);
  while Length(Result) < 3 do
    Result := '0' + Result;
end;

function  TformProgramManager.numberToFileName(base: string; number: Integer): string;
begin
  Result := ChangeFileExt(base, '.' + generateFileExt(number));
end;

function  TformProgramManager.indexToFileName(base: string; index: Integer): string;
begin
  Result := numberToFileName(base, StrToInt(formLightning.programList[index]));
end;

function  TformProgramManager.determineFreeProgramNumber(): Integer;
{
  Pre:  desktop <> nil;
  Post: Result = a number greater than or equal to 0, yet as low as possible,
                 that is not present in the list of program files yet; 
}
begin
  Result := 0;
  while formLightning.programList.IndexOf(IntToStr(Result)) > -1 do
    Inc(Result);
end;

function  TformProgramManager.requireDesktopSaved(noConfirm: Boolean): Boolean;
begin
  Result :=
     (
      // the desktop was already saved before
      (formLightning.currentFile <> '')
      or
      // the desktop was never saved, request the user to save it
      (MessageDlg('To continue, the current desktop must be saved.'#13#10'Would you like to continue?', mtConfirmation, [mbYes, mbNo], 0) = mrYes)
     )
     and
     (
      noConfirm
      or
      disableConfirm
      or
      // the desktop was already saved before, but may have changes
      ((formLightning.currentFile <> '') and (MessageDlg('To continue, any changes in the current program must be saved.'#13#10'Would you like to continue?', mtConfirmation, [mbYes, mbNo], 0) = mrYes))
     );
end;

function  TformProgramManager.saveCurrentProgram(noConfirm: Boolean): Boolean;
begin
  // before the program can be saved, the desktop must be saved
  if requireDesktopSaved(noConfirm) then
  begin
    if formLightning.currentFile = '' then
      Result := formLightning.saveDialog()
    else
    begin
      formSceneManager.saveProgram(
        formProgramManager.indexToFileName(formLightning.currentFile, currentProgramIndex),
        currentProgramName
      );
      Result := True;
    end;
  end
  else
    Result := False;
end;

function  TformProgramManager.newProgram(noConfirm: Boolean): Boolean;
{ YET TO BE CAREFULLY TESTED!!! }
var
  i: Integer;
begin
  Result := False;
  formLightning.stopTimer();
  try
  {if noConfirm or
     disableConfirm or
     ((currentFile = '') and (MessageDlg('In order to be able to continue, the current desktop must be saved.'#13#10'Would you like to continue?', mtConfirmation, [mbYes, mbNo], 0) = mrYes))
     or
     ((currentFile <> '') and (MessageDlg('In order to be able to continue, any changes in the current desktop and program must be saved.'#13#10'Would you like to continue?', mtConfirmation, [mbYes, mbNo], 0) = mrYes))
     then}
    if saveCurrentProgram(False) then
    begin
      Result := True;
      // find an unused program number that can be used as program file extension number.
      i := determineFreeProgramNumber();
      formLightning.programList.Add(IntToStr(i));
      lbPrograms.Items.Add('Program');

      // create an empty program and store it to disk,
      // using the program extension number that was just determined.
      with TStringList.Create do
      begin
        try
          Add('Program|0');
          SaveToFile(numberToFileName(formLightning.currentFile, i)); // this line was bugging in an earlier version!!!
        finally
          Free();
        end;
      end;
      formLightning.saveProgramList(formLightning.currentFile);
      //fLightning.miSaveClick(fProgramManager);
    end;
  finally
    formLightning.startTimer();
  end;
end;

{function  TfProgramManager.saveProgram(noConfirm: Boolean): Boolean;
begin
  //Result := False;
  //if MessageDlg('In order to be able to continue, the current program and desktop must be saved.'#13#10'Would you like to continue?', mtConfirmation, [mbYes, mbNo], 0) = mrYes then
  //begin
  Result := False;
  if noConfirm or
     disableConfirm or
     ((currentFile = '') and (MessageDlg('In order to be able to continue, the current desktop must be saved.'#13#10'Would you like to continue?', mtConfirmation, [mbYes, mbNo], 0) = mrYes))
     or
     ((currentFile <> '') and (MessageDlg('In order to be able to continue, any changes in the current desktop and program must be saved.'#13#10'Would you like to continue?', mtConfirmation, [mbYes, mbNo], 0) = mrYes))
     then
  begin
    fLightning.miSaveClick(fProgramManager);
    //fSceneManager.saveProgram(ChangeFileExt(ExtractFileName(currentFile), '.' + generateFileExt(StrToInt(desktop[currentProgram]))), currentProgramName);
    Result := True;
  end;
  //end;
end;}

function  TformProgramManager.switchProgram(index: Integer; noConfirm: Boolean): Boolean;
begin
  Result := False;
  if index = currentProgramIndex then
    Result := True
  else
  begin
    formLightning.stopTimer();
    try
      if {saveProgram(noConfirm)}saveCurrentProgram(noConfirm) then
      begin
        Result := True;
        loadProgram(index);
        //currentProgram := i;
        //fSceneManager.loadProgram(generateFileNameFromIndex(currentFile, i));
        //fSceneManager.loadProgram(ChangeFileExt(ExtractFileName(currentFile), '.' + generateFileExt(StrToInt(desktop[i]))));
      end;
    finally
      formLightning.startTimer();
    end;
  end;
end;

procedure TformProgramManager.loadProgram(index: Integer);
begin
  currentProgramIndex  := index;
  lbPrograms.ItemIndex := index;
  formSceneManager.loadProgram(indexToFileName(formLightning.currentFile, index));
end;

function  TformProgramManager.getProgramName(index: Integer): string;
var
  bString: string;
begin
  if index = currentProgramIndex then
  begin
    // return program name from memory
    Result := currentProgramName
  end
  else
  begin
    // read program name from disk
    with TStringList.Create() do
    begin
      LoadFromFile(indexToFileName(formLightning.currentFile, index));
      bString := Strings[0];
      Result  := parse(bString, '|');
      Free();
    end;
  end;
end;

procedure TformProgramManager.FormClose(Sender: TObject;
  var Action: TCloseAction);
begin
  Action := caNone;
  formLightning.miProgramManagerClick(Sender);
end;

procedure TformProgramManager.miApplyToAllProgramsClick(Sender: TObject);
var
  i:       Integer;
  bString: string;
  fileName: string;
begin
  if formLightning.programList.Count > 1 then
    for i := 0 to formLightning.programList.Count - 1 do
      if i <> currentProgramIndex then
      begin
        // change the fade time in all programs on disk.
        with TStringList.Create do
        begin
          try
            fileName := indexToFileName(formLightning.currentFile, i);
            LoadFromFile(fileName);
            bString := Strings[0];
            Strings[0] := parse(bString, '|') + '|' + IntToStr(upDownFadeTime.Position);
            SaveToFile(fileName);
          finally
            Free();
          end;
        end;
      end;
end;

procedure TformProgramManager.tbNewClick(Sender: TObject);
begin
  newProgram(False);
end;

procedure TformProgramManager.tbOpenClick(Sender: TObject);
var
  i: Integer;
begin
  i := lbPrograms.ItemIndex;
  if i > -1 then
    switchProgram(i, False);
end;

procedure TformProgramManager.tbRenameClick(Sender: TObject);
var
  i: Integer;
  name, bString: string;
  fileName: string;
begin
  i := lbPrograms.ItemIndex;
  if i > -1 then
  begin
    name := InputBox('Rename', 'Please enter new program name:', lbPrograms.Items[i]);
    lbPrograms.Items[i] := name;
    if i = currentProgramIndex then
    begin
      // change program name in memory.
      currentProgramName := name
    end
    else
    begin
      // change program name in program on disk.
      with TStringList.Create do
      begin
        fileName := indexToFileName(formLightning.currentFile, i);
        LoadFromFile(fileName);
        bString := Strings[0];
        parse(bString, '|'); // bString now contains the additional program information
        Strings[0] := name + '|' + bString;
        SaveToFile(fileName);
        Free();
      end;
    end;
  end;
end;

procedure TformProgramManager.tbDuplicateClick(Sender: TObject);
{ YET TO BE CAREFULLY TESTED!!! }
var
  sourceIndex, destNumber: Integer;
begin
  formLightning.stopTimer();
  try
    sourceIndex := lbPrograms.ItemIndex;
    if (sourceIndex > -1) then
    begin
      // before copying the current program, make sure any changes to it are saved.
      if saveCurrentProgram(False) then
      begin
        destNumber := determineFreeProgramNumber();
        formLightning.programList.Add(IntToStr(destNumber));
        lbPrograms.Items.Add(getProgramName(sourceIndex));
        CopyFile(
          PChar(indexToFileName(formLightning.currentFile, sourceIndex)),
          PChar(numberToFileName(formLightning.currentFile, destNumber)),
          False);
        formLightning.saveProgramList(formLightning.currentFile);
      end;
    end;
  finally
    formLightning.startTimer();
  end;
end;

procedure TformProgramManager.tbDeleteClick(Sender: TObject);
var
  i: Integer;
begin
  formLightning.stopTimer();
  try
    i := lbPrograms.ItemIndex;
    if i > -1 then
    begin
      //if desktop.Count > 1 then
      //begin
      if (MessageDlg('The selected program will be permantently deleted'#13#10'and any changes to the current program will be saved.'#13#10'Are you sure?', mtConfirmation, [mbYes, mbNo], 0) = mrYes)
         and
         requireDesktopSaved(True) then
      begin
        // make sure that a program is opened, other than the one that will be deleted.
        if i = currentProgramIndex then
        begin
          if formLightning.programList.Count = 1 then
          begin
            // if this is the last program that's present,
            // create an empty new program first.
            newProgram(True);
            switchProgram(1, True);
          end
          else if i = 0 then
          begin
            // if this is the first program in the list, open the second program.
            switchProgram(1, True) // !!needs more checking!!
          end
          else
          begin
            // if this isn't the first program in the list, open the prior program.
            switchProgram(i - 1, True);
          end;
        end;
        // delete the program from disk.
        DeleteFile(indexToFileName(formLightning.currentFile, i));

        lbPrograms.Items.Delete(i);
        formLightning.programList.Delete(i);
        if currentProgramIndex > i then // as the current program comes after the program that is to be deleted from the desktop list
          Dec(currentProgramIndex); // the index of the current program has changed
        //fLightning.miSaveClick(fProgramManager);
        formLightning.saveProgramList(formLightning.currentFile);
      end;
      //end;
      //else
      //  MessageBox(Handle, PChar('At least one program is required to exist at all times.'#13#10'In order to delete this program, please create a new program prior to deleting this.'), 'Information', MB_ICONEXCLAMATION);
    end;
  finally
    formLightning.startTimer();
  end;
end;

procedure TformProgramManager.tbUpClick(Sender: TObject);
var
  i: Integer;
  bString: string;
begin
  formLightning.stopTimer();
  try
    i := lbPrograms.ItemIndex;
    if i > 0 then
    begin
      // swap program numbers
      bString := formLightning.programList[i - 1];
      formLightning.programList[i - 1] := formLightning.programList[i];
      formLightning.programList[i] := bString;
      // swap program titles in list box
      bString := lbPrograms.Items[i - 1];
      lbPrograms.Items[i - 1] := lbPrograms.Items[i];
      lbPrograms.Items[i] := bString;
      if i = currentProgramIndex then
        // shift currentProgramIndex
        Dec(currentProgramIndex);
      // shift list box selection to the program's new index
      lbPrograms.ItemIndex := lbPrograms.ItemIndex - 1;
      formLightning.saveProgramList(formLightning.currentFile);
    end;
  finally
    formLightning.startTimer();
  end;
end;

procedure TformProgramManager.tbDownClick(Sender: TObject);
var
  i: Integer;
  bString: string;
begin
  formLightning.stopTimer();
  try
    i := lbPrograms.ItemIndex;
    if (i > -1) and (i < lbPrograms.Count - 1) then
    begin
      // swap program numbers
      bString := formLightning.programList[i];
      formLightning.programList[i] := formLightning.programList[i + 1];
      formLightning.programList[i + 1] := bString;
      // swap program titles in list
      bString := lbPrograms.Items[i];
      lbPrograms.Items[i] := lbPrograms.Items[i + 1];
      lbPrograms.Items[i + 1] := bString;
      if i = currentProgramIndex then
        // shift currentProgramIndex
        Inc(currentProgramIndex);
      // shift list box selection to the program's new index
      lbPrograms.ItemIndex := lbPrograms.ItemIndex + 1;
      formLightning.saveProgramList(formLightning.currentFile);
    end;
  finally
    formLightning.startTimer();
  end;
end;

end.
