unit unitmodule;

(* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1
 *
 * The contents of this file are subject to the Mozilla Public License Version 
 * 1.1 (the "License"); you may not use this file except in compliance with 
 * the License. You may obtain a copy of the License at 
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is the Lightning TModule class.
 *
 * The Initial Developer of the Original Code is
 * Tomsoft.
 * Portions created by Tomsoft are Copyright (C) 2009
 * Tomsoft. All Rights Reserved.
 *
 * Contributor(s):
 *   Tommy Sools <tom@sools.com>
 *
 * ***** END LICENSE BLOCK ***** *)

{$DEFINE VIEWCOMPONENT}

interface

uses
  Forms, Dialogs, SysUtils, Windows, Messages,
  Graphics, StdCtrls, Controls, ExtCtrls, Classes, Math, Xdom_3_2, common,
  unitconnector;

type
  TModuleColor = class(TShape);
  TModuleLayer = class(TLabel);
  TModuleKind = (mkNormal, mkComponent, mkComponentIn, mkComponentOut, mkLua, mkDLOP);
  TModule = class(TPanel)
  private
    iUsedForScene: TImage;
    layer: Integer;
    inputConnected, outputConnected: string;
    constructor Create(Parent: TWinControl; ext, sceneid: string); reintroduce;
    procedure moduleMouseDown(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
    procedure moduleMouseUp(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
    procedure moduleStartDrag(Sender: TObject; var DragObject: TDragObject);
    procedure moduleEndDrag(Sender, Target: TObject; X, Y: Integer);
    procedure iUsedForSceneMouseDown(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
  public
    instance:  Integer;
    lInfo:     TLabel;
    input:     array[1..maxChannels] of TInput;
    output:    array[1..maxChannels] of TOutput;
    usedForScene: Boolean;
    ext:       string;
    kind:      TModuleKind;
    sceneid:   string;
    excludeFromFadeTime: Boolean;
    sColor:    TModuleColor;
    labelLayer: TLabel;
    iSettings: TImage;
    module:    PLightningModule;
    destroyM:  TModuleDestroy;
    action:    TModuleAction;
    show:      TModuleShow;
    hide:      TModuleHide;
    open:      TModuleOpen;
    save:      TModuleSave;
    openscene: TModuleOpenScene;
    savescene: TModuleSaveScene;
    keyboard:  TModuleKeyboard;
    class function createNew(Parent: TWinControl; top, left: Integer; kind, sceneid: string; visible: Boolean): TModule;
    class function createFromXML(xmlNode: TDOMNode; Parent: TWinControl; offsetTop, offsetLeft: Integer; mapModuleId: TStringList; isSelected: Boolean; visible: Boolean): TModule;
    procedure addConnectionsFromXML(xmlNode: TDOMNode; mapModuleId: TStringList);
    class function createFromString(s: string; Parent: TWinControl; offsetTop, offsetLeft: Integer; isSelected: Boolean; visible: Boolean): TModule;
    destructor Destroy(); override;
    procedure addToModuleArray();
    procedure addToSelectedModuleArray();
    procedure removeFromModuleArray();
    procedure removeFromSelectedModuleArray();
    procedure closeDetails(freeDetails: Boolean);
    procedure refreshOutputConnected();
    procedure refreshInputConnected();
    procedure preventCircle(checkModule: TModule);
    procedure moduleMouseMove(Sender: TObject; Shift: TShiftState; X, Y: Integer);
    procedure lInfoClick(Sender: TObject);
    function toXML(): string;
    function toString(includeId: Boolean): string;
    function getColor(): TColor;
    procedure setColor(color: TColor);
    function getCaption(): string;
    procedure setCaption(c: string);
    procedure updateWidth();
    function getIsUsedForScene(): Boolean;
    procedure setIsUsedForScene(usedForScene: Boolean);
    procedure setIsSelected(selected: Boolean);
    function getIsSelected(): Boolean;
    procedure updateLayer();
    procedure setLayer(layer: Integer);
    function getLayer(): Integer;
    procedure setLocked(locked: Boolean);
    procedure BringToFront(); reintroduce;
    procedure SendToBack(); reintroduce;
    procedure actionOutput(input, output: Integer; bridge: TSpectrum);
    procedure actionUnconnectedInputs(input: Integer);
    procedure actionInput(input: Integer; var bridge: TSpectrum; checkLUA: Boolean);
  end;
  TModuleArray = array of TModule;

implementation

uses
  unitlightningsettings, unitformlightning, unitformdetails,
  unitmodulecomponent, unitmodulecomponentin, unitmodulecomponentout,
  unitmodulelua, unitmoduledlop,
  unitdesktopview, unitformdlop, unitformtrace;

constructor TModule.Create(Parent: TWinControl; ext, sceneid: string);
var
  bExternal:          PExternal;
  createM:            TModuleCreate;
  tooltip:            string;
  bPLightningModule2: PLightningModule2;
  bPLightningModule3: PLightningModule3;
  i, minWidth, minHeight, inputTop, outputTop: Integer;
  desktopView: TDesktopView;
begin
  inherited Create(Owner);

  inputConnected   := '';
  outputConnected  := '';

  //Color           := $00F3FFE8;//$00D5FFB1;//$00FFFEF0;
  ParentBackground := False;
  Color            := clBtnFace;

  Visible         := False;
  Self.Parent     := Parent;

  OnMouseMove     := moduleMouseMove;
  OnMouseDown     := moduleMouseDown;
  OnMouseUp       := moduleMouseUp;
  OnStartDrag     := moduleStartDrag;
  OnEndDrag       := moduleEndDrag;
  OnDragOver      := TDesktopView(Parent).OnDragOver;
  OnDragDrop      := TDesktopView(Parent).OnDragDrop;
  DragMode        := dmManual;

  Self.ext            := ext;
  Self.sceneid        := sceneid;
  excludeFromFadeTime := False;
  layer               := 0;

  bExternal := locateExternal(ext);
  kind := bExternal.moduleKind;
  openscene := nil;
  savescene := nil;
  keyboard  := nil;
  case kind of
    mkNormal: // .dll (normal)
    begin
      createM   := GetProcAddress(bExternal.hInstance, 'create');
      destroyM  := GetProcAddress(bExternal.hInstance, 'destroy');
      action    := GetProcAddress(bExternal.hInstance, 'action');
      show      := GetProcAddress(bExternal.hInstance, 'show');
      hide      := GetProcAddress(bExternal.hInstance, 'hide');
      open      := GetProcAddress(bExternal.hInstance, 'open');
      save      := GetProcAddress(bExternal.hInstance, 'save');
      openscene := GetProcAddress(bExternal.hInstance, 'openscene');
      savescene := GetProcAddress(bExternal.hInstance, 'savescene');
      keyboard  := GetProcAddress(bExternal.hInstance, 'keyboard');
      module    := createM(Handle);
    end;
    mkComponent: // .lig
    begin
      destroyM  := unitmodulecomponent.destroy;
      action    := unitmodulecomponent.action;
      show      := unitmodulecomponent.show;
      hide      := unitmodulecomponent.hide;
      open      := unitmodulecomponent.open;
      save      := unitmodulecomponent.save;
      desktopView := TDesktopView.Create(Self);
      with desktopView do
      begin
        Parent := Self;
        addFromXMLFile(bExternal.filePath + bExternal.fileName, 'lightningdesktop');
        Visible := False;
        {$IFDEF VIEWCOMPONENT}
        Self.Height := 500;
        Align := alNone;
        Top := 100;
        Height := 380;
        Width := 380;
        Visible := True;
        {$ENDIF}
      end;
      module := unitmodulecomponent.create(Handle, bExternal.filePath + bExternal.fileName, desktopView);
    end;
    mkComponentIn:
    begin
      destroyM  := unitmodulecomponentin.destroy;
      action    := unitmodulecomponentin.action;
      show      := unitmodulecomponentin.show;
      hide      := unitmodulecomponentin.hide;
      open      := unitmodulecomponentin.open;
      save      := unitmodulecomponentin.save;
      if TDesktopView(Parent).Parent is TModule then
        desktopView := Pointer(TDesktopView(Parent))
      else
        desktopView := nil;
      module    := unitmodulecomponentin.create(Handle, desktopView);
    end;
    mkComponentOut:
    begin
      destroyM  := unitmodulecomponentout.destroy;
      action    := unitmodulecomponentout.action;
      show      := unitmodulecomponentout.show;
      hide      := unitmodulecomponentout.hide;
      open      := unitmodulecomponentout.open;
      save      := unitmodulecomponentout.save;
      if TDesktopView(Parent).Parent is TModule then
        desktopView := Pointer(TDesktopView(Parent))
      else
        desktopView := nil;
      module    := unitmodulecomponentout.create(Handle, desktopView);
    end;
    mkLUA: // .lua
    begin
      destroyM  := unitmodulelua.destroy;
      action    := unitmodulelua.action;
      show      := unitmodulelua.show;
      hide      := unitmodulelua.hide;
      open      := unitmodulelua.open;
      save      := unitmodulelua.save;
      module    := unitmodulelua.create(Handle, bExternal.filePath + bExternal.fileName);
    end;
    mkDLOP: // .dll (DLOP)
    begin
      destroyM  := unitmoduledlop.destroy;
      action    := unitmoduledlop.action;
      show      := unitmoduledlop.show;
      hide      := unitmoduledlop.hide;
      open      := unitmoduledlop.open;
      save      := unitmoduledlop.save;
      module    := unitmoduledlop.create(Handle, bExternal.filePath + bExternal.fileName);
    end;
  end;
  module^.openFile := openFile;
  if module^.reserved <> nil then
  begin
    bPLightningModule2              := PLightningModule2(module^.reserved);
    bPLightningModule2^.scenic      := True;
    bPLightningModule2^.nextScene   := nextScene;
    bPLightningModule2^.nextProgram := nextProgram;
    bPLightningModule2^.setFadeTime := setFadeTime;
    if bPLightningModule2^.reserved <> nil then
      PLightningModule3(bPLightningModule2^.reserved)^.nextProgramEx := nextProgramEx;
  end;

  lInfo := TLabel.Create(Self);
  with lInfo do
  begin
    Font.Name  := 'MS Sans Serif';
    Font.Size  := 8;
    Font.Style := lInfo.Font.Style + [fsBold];
    Font.Color := $00533200;//$00676766;
    if module^.infoTop > -1 then
      Top := module^.infoTop
    else
      Top := 8;
    if not(lightningSettings.lock) then
    begin
      Cursor  := crHandPoint;
      OnClick := lInfoClick;
    end;
    setCaption(module^.title);
    OnDragOver := TDesktopView(Self.Parent).OnDragOver;
    OnDragDrop := TDesktopView(Self.Parent).OnDragDrop;
    Parent := Self;
  end;
  (*if module^.settings then
  begin
    info.Font.Color := $002380FF;
    //$002179F2;//$002A4500;//$00345500;//$00426C00;//$0068AA00;//$0000B84C;//$0072572E;//$008B5400;
    info.Cursor     := crHandPoint;
    info.OnClick    := fLightning.lExternalClick;
  end
  //else if {sceneid <> ''}@openscene <> nil then
  //  info.Font.Color := $000000FF
  else*)
  if module^.inputs > 0 then
  begin
    if module^.infoTop = -1 then
      lInfo.Top := lInfo.Top + 24;
    tooltip := module^.inputTooltip;
    if module^.inputTop > -1 then
      inputTop := module^.inputTop
    else
      inputTop := 8;
    for i := 1 to module^.inputs do
    begin
      input[i] := TInput.Create(Self);
      with input[i] do
      begin
        number   := i;
        Left     := 8 + (i - 1) * 24;
        Hint     := parse(tooltip, '|');
        Parent   := Self;
        Top      := inputTop;
      end;
      input[i].OnDragDrop := TDesktopView(Parent).OnDragDrop;
    end;
  end;
  minHeight := lInfo.Top + 24;
  if module^.outputs > 0 then
  begin
    tooltip := module^.outputTooltip;
    if module^.outputTop > -1 then
      outputTop := module^.outputTop
    else
      outputTop := lInfo.Top + 24;
    for i := 1 to module^.outputs do
    begin
      output[i] := TOutput.Create(Self);
      with output[i] do
      begin
        number   := i;
        Top      := outputTop;
        Hint     := parse(tooltip, '|');
        Parent   := Self;
      end;
      output[i].OnDragDrop := TDesktopView(Parent).OnDragDrop;
    end;
    minHeight := outputTop + 24;
  end;
  if module^.Height > minHeight then
    Height := module^.height
  else
    Height := minHeight;
  if @openscene <> nil then
  begin
    usedForScene := False;
    iUsedForScene := TImage.Create(Self);
    with iUsedForScene do
    begin
      Cursor      := crHandPoint;
      AutoSize    := True;
      Hint        := 'Use for scene';
      ShowHint    := True;
      Picture.Bitmap.LoadFromResourceName(hInstance, 'usedforsceneup');
      Top         := Self.Height - Height - 1;
      OnMouseDown := iUsedForSceneMouseDown;
      OnDragOver  := TDesktopView(Self.Parent).OnDragOver;
      OnDragDrop  := TDesktopView(Self.Parent).OnDragDrop;
      Parent      := Self;
    end;
  end
  else
    iUsedForScene := nil;
  sColor := TModuleColor.Create(Self);
  with sColor do
  begin
    Top         := 1;
    Left        := 1;
    Height      := 5;
    Brush.Color := clBtnFace;
    Pen.Color   := clBtnFace;
    OnMouseDown := moduleMouseDown;
    OnMouseUp   := moduleMouseUp;
    OnStartDrag := moduleStartDrag;
    OnEndDrag   := moduleEndDrag;
    DragMode    := dmManual;
    OnDragOver  := TDesktopView(Self.Parent).OnDragOver;
    OnDragDrop  := TDesktopView(Self.Parent).OnDragDrop;
    Parent      := Self;
  end;
  labelLayer := TModuleLayer.Create(Self);
  with labelLayer do
  begin
    Top         := 0;
    Left        := 2;
    Font.Name   := 'Arial';
    Font.Size   := 8;
    Font.Color  := clGray;
    Font.Style  := Font.Style + [fsBold];
    Caption     := '';
    OnMouseDown := moduleMouseDown;
    OnMouseUp   := moduleMouseUp;
    OnStartDrag := moduleStartDrag;
    OnEndDrag   := moduleEndDrag;
    DragMode    := dmManual;
    OnDragOver  := TDesktopView(Self.Parent).OnDragOver;
    OnDragDrop  := TDesktopView(Self.Parent).OnDragDrop;
    Parent      := Self;
  end;
  if module^.settings then
  begin
    iSettings := TImage.Create(Self);
    with iSettings do
    begin
      Transparent := True;
      AutoSize    := True;
      Picture.Bitmap.LoadFromResourceName(hInstance, 'settings');
      Top        := 1;
      Cursor     := crHandPoint;
      OnMouseUp  := formLightning.iSettingsMouseUp;
      OnDragOver := TDesktopView(Self.Parent).OnDragOver;
      OnDragDrop := TDesktopView(Self.Parent).OnDragDrop;
      Parent     := Self;
    end;
  end
  else
    iSettings := nil;
  updateWidth();
  addToModuleArray();
end;

class function TModule.createNew(Parent: TWinControl; top, left: Integer; kind, sceneid: string; visible: Boolean): TModule;
var
  desktopView: TDesktopView;
  module: TModule;
begin
  Result := nil;
  desktopView := TDesktopView(Parent);
  formLightning.stopTimer();
  try
    module := desktopView.findModuleName(kind);
    if (module = nil) or (module.kind <> mkDLOP) then  
    begin
      Result := TModule.Create(Parent, kind, sceneid);
      Result.SetBounds(left, top, Result.Width, Result.Height);
      Result.Visible := visible;
    end
    else
    begin
      raise Exception.Create(
        'This desktop already contains a DLOP of type "' + kind + '".'#13#10 +
        'It is not possible to add multiple DLOPs of the same type to a desktop.'
      );
    end;
  finally
    formLightning.startTimer();
  end;
end;

class function TModule.createFromXML(xmlNode: TDOMNode; Parent: TWinControl; offsetTop, offsetLeft: Integer; mapModuleId: TStringList; isSelected: Boolean; visible: Boolean): TModule;
  {
    Pre: text uses Linux line end characters (#10);
    Post: Result = line end characters in text were converted to
          Windows line end characters (#13#10) and
          each line's leading and trailing whitespaces were trimmed;
  }
  function trimXML(text: string): string;
  var
    searchPosition, p, l: Integer;
  begin
    Result := '';
    l := Length(text);
    searchPosition := 1;
    repeat
      p := Pos(#10, Copy(text, searchPosition, l));
      if p = 0 then
        p := l - searchPosition + 2;
      begin
        Result := Result + Trim(Copy(text, searchPosition, p - 1)) + #13#10;
        Inc(searchPosition, p);
      end;
    until searchPosition > l;
    Result := Trim(Result);
  end;
var
  xmlAttrNode: TDOMNode;
  bModule: TModule;
  top, left, layer: Integer;
  fadeTime: Boolean;
  color: TColor;
  bString, generatedSceneid, sceneid, ext, title, settings: string;
begin
  sceneid := '';
  ext := '';
  top := 0;
  left := 0;
  fadeTime := False;
  layer := 0;
  color := clBtnFace;
  title := '';
  settings := '';
  // lees de attributen
  xmlAttrNode := xmlNode.Attributes.GetNamedItem('id');
  if xmlAttrNode <> nil then
    sceneid := xmlAttrNode.TextContent;
  xmlAttrNode := xmlNode.Attributes.GetNamedItem('top');
  if xmlAttrNode <> nil then
    top := StrToIntDef(xmlAttrNode.TextContent, 0);
  xmlAttrNode := xmlNode.Attributes.GetNamedItem('left');
  if xmlAttrNode <> nil then
    left := StrToIntDef(xmlAttrNode.TextContent, 0);
  xmlAttrNode := xmlNode.Attributes.GetNamedItem('layer');
  if xmlAttrNode <> nil then
    layer := StrToIntDef(trimXML(xmlAttrNode.TextContent), 0);
  // doorloop alle elementen binnen het XML element
  xmlNode := xmlNode.FindFirstChildElement();
  while xmlNode <> nil do
  begin
    if xmlNode.NodeName = 'file' then
      ext := trimXML(xmlNode.TextContent)
    else if xmlNode.NodeName = 'fadetime' then
      fadeTime := True
    else if xmlNode.NodeName = 'color' then
    begin
      bString := trimXML(xmlNode.TextContent);
      color :=
        RGB(
          StrToIntDef('$' + Copy(bString, 2, 2), 0),
          StrToIntDef('$' + Copy(bString, 4, 2), 0),
          StrToIntDef('$' + Copy(bString, 6, 2), 0)
        );
    end
    else if xmlNode.NodeName = 'title' then
      title := trimXML(xmlNode.TextContent)
    else if xmlNode.NodeName = 'settings' then
      settings := trimXML(xmlNode.TextContent);
    xmlNode := xmlNode.FindNextSiblingElement();
  end;
  if locateExternal(ext) <> nil then
  begin
    if (sceneid = '') or (mapModuleId <> nil) then
    begin
      generatedSceneId := TDesktopView(Parent).generateSceneId();
      if mapModuleId <> nil then
      begin
        // keep track of which original module id corresponds to
        // which generated module id, so that connections can be restored
        // after pasting from clipboard.
        mapModuleId.AddObject(generatedSceneId, TObject(StrToIntDef(sceneId, 0)));
      end;
      sceneid := generatedSceneId;
    end;
    bModule := createNew(Parent, top + offsetTop, left + offsetLeft, ext, sceneid, visible);
    bModule.setIsSelected(isSelected);
    bModule.excludeFromFadeTime := not(fadeTime);
    bModule.setLayer(layer);
    if color <> clBtnFace then
      bModule.setColor(color);
    if title <> '' then
    begin
      bModule.setCaption(title);
      bModule.updateWidth();
    end;
    bModule.open(bModule.module, PChar(settings));
    Result := bModule;
  end
  else
  begin
    raise Exception.Create(
      'This desktop depends on the "' + ext + '" extension. Please'#13#10 +
      'make sure it is available, restart the program and try again.'
    );
  end;
end;

procedure TModule.addConnectionsFromXML(xmlNode: TDOMNode; mapModuleId: TStringList);
  {
    Pre: text uses Linux line end characters (#10);
    Post: Result = line end characters in text were converted to
          Windows line end characters (#13#10) and
          each line's leading and trailing whitespaces were trimmed;
  }
  function trimXML(text: string): string;
  var
    searchPosition, p, l: Integer;
  begin
    Result := '';
    l := Length(text);
    searchPosition := 1;
    repeat
      p := Pos(#10, Copy(text, searchPosition, l));
      if p = 0 then
        p := l - searchPosition + 2;
      begin
        Result := Result + Trim(Copy(text, searchPosition, p - 1)) + #13#10;
        Inc(searchPosition, p);
      end;
    until searchPosition > l;
    Result := Trim(Result);
  end;
var
  xmlAttrNode: TDOMNode;
  number: Integer;
begin
  // walk thru all elements inside the XML element
  xmlNode := xmlNode.FindFirstChildElement();
  while xmlNode <> nil do
  begin
    if xmlNode.NodeName = 'output' then
    begin
      number := 0;
      xmlAttrNode := xmlNode.Attributes.GetNamedItem('number');
      if xmlAttrNode <> nil then
        number := StrToIntDef(xmlAttrNode.TextContent, 0);
      if number > 0 then
      begin
        if number <= module^.outputs then
          output[number].connectFromXML(xmlNode, mapModuleId)
        else
          raise Exception.Create('There was an error reading the XML file. The module contains specifications for more output links than exist in the module.');
      end;
    end;
    xmlNode := xmlNode.FindNextSiblingElement();
  end;
end;

class function TModule.createFromString(s: string; Parent: TWinControl; offsetTop, offsetLeft: Integer; isSelected: Boolean; visible: Boolean): TModule;
var
  bModule: TModule;
  bString: string;
  layer: Integer;
  excludeFromFadeTime: Boolean;
  originalExt, ext, top, left, sceneid, settings, options, color, title: string;
begin
  parse(s, '|', originalExt);
  parse(s, '|', top);
  parse(s, '|', left);
  parse(s, '|', sceneid);
  excludeFromFadeTime := parse(s, '|') <> '';
  layer := StrToIntDef(parse(s, '|'), 0);
  parse(s, '|', options);
  parse(s, '|');
  parse(s, '|', settings);
  // compatibility with saved desktops that use removed
  // inStereoAudio, outZoomLeftSpectrum, outZoomRightSpectrum, procLUA,
  // outLargeLabel or outDLOP modules.
  if originalExt = 'in_stereoaudio.dll' then
    ext := 'in_audio.dll'
  else if originalExt = 'proc_lua.dll' then
  begin
    bString := ExtractFileName(settings);
    if FileExists(ExtractFilePath(ParamStr(0)) + 'lua\' + bString) then
    begin
      ext := bString;
      settings := '';
    end
    else
      ext := originalExt;
  end
  else if originalExt = 'out_zoomleftspectrum.dll' then
    ext := 'out_spectrum.dll'
  else if originalExt = 'out_zoomrightspectrum.dll' then
    ext := 'out_spectrum.dll'
  else if originalExt = 'out_largelabel.dll' then
    ext := 'out_label.dll'
  else if originalExt = 'out_dlop.dll' then
  begin
    if formDLOP.ShowModal() = mrOk then
      ext := formDLOP.getFileName()
    else
      ext := originalExt;
  end
  else
    ext := originalExt;
(*
  if originalExt = 'out_multisoftlight.dll' then
  begin
    // replace outMultiRGBLight with 8 x outRGBLight
    for i := 7 downto 0 do
      Result := TModule.createFromString('out_softlight.dll|' + s, Parent, offsetTop, offsetLeft + (i * {81}72), isSelected);
  end
  else *)if locateExternal(ext) <> nil then
  begin
    if Trim(sceneid) = '' then
      sceneid := TDesktopView(Parent).generateSceneid();
    bModule := createNew(TDesktopView(Parent), StrToIntDef(top, 10) + offsetTop, StrToIntDef(left, 10) + offsetLeft, ext, sceneid, visible);
    bModule.setIsSelected(isSelected);
    bModule.excludeFromFadeTime := excludeFromFadeTime;
    bModule.setLayer(layer);
    color := Trim(parse(options, ','));
    if color <> '' then
      bModule.setColor(StrToIntDef(color, clBtnFace));
    title := Trim(parse(options, ','));
    // compatibility with saved desktops that use former
    // inStereoAudio, outZoomLeftSpectrum or outZoomRightSpectrum modules.
    if originalExt = 'in_stereoaudio.dll' then
      settings := ',4096'
    else if originalExt = 'out_zoomleftspectrum.dll' then
      settings := 'l'
    else if originalExt = 'out_zoomrightspectrum.dll' then
      settings := 'r';
    // compatibility with saved desktops that use old version of
    // outLabel or outLargeLabel modules.
    if (title = '') and (ext = 'out_label.dll') then
    begin
      bString := settings;
      title := Trim(parse(bString, ','));
    end;
    if title <> '' then
    begin
      bModule.setCaption(title);
      bModule.updateWidth();
    end;
    bModule.open(bModule.module, PChar(settings));
    Result := bModule;
  end
  else
  begin
    raise Exception.Create(
      'This desktop depends on the "' + ext + '" extension. Please'#13#10 +
      'make sure it is available, restart the program and try again.'
    );
  end;
end;

destructor TModule.Destroy();
var
  i: Integer;
begin
  closeDetails(True);
  // disconnect and release all inputs and outputs
  for i := 1 to module^.inputs do
    input[i].disconnect();
  for i := 1 to module^.inputs do
    input[i].Free();
  for i := 1 to module^.outputs do
    output[i].disconnect();
  for i := 1 to module^.outputs do
    output[i].Free();
  lInfo.Free();
  iUsedForScene.Free();
  labelLayer.Free();
  sColor.Free();
  if iSettings <> nil then
    iSettings.Free();
  destroyM(module);
  removeFromModuleArray();
  inherited Destroy;
end;

procedure TModule.addToModuleArray();
var
  desktopView: TDesktopView;
  i: Integer;
begin
  desktopView := TDesktopView(Parent);
  // check if the module doesn't already exist in the array
  for i := 0 to Length(desktopView.modules) - 1 do
    if desktopView.modules[i] = Self then
      Exit;
  SetLength(desktopView.modules, Length(desktopView.modules) + 1);
  desktopView.modules[Length(desktopView.modules) - 1] := Self;
end;

procedure TModule.addToSelectedModuleArray();
var
  desktopView: TDesktopView;
  i: Integer;
begin
  desktopView := TDesktopView(Parent);
  // check if the module doesn't already exist in the array
  for i := 0 to Length(desktopView.selectedModules) - 1 do
    if desktopView.selectedModules[i] = Self then
      Exit;
  SetLength(desktopView.selectedModules, Length(desktopView.selectedModules) + 1);
  desktopView.selectedModules[Length(desktopView.selectedModules) - 1] := Self;
end;

procedure TModule.removeFromModuleArray();
var
  desktopView: TDesktopView;
  i, c, index: Integer;
begin
  // detect the position in the module array of the module that will be deleted
  desktopView := TDesktopView(Parent);
  c := Length(desktopView.modules);
  index := -1;
  for i := 0 to c - 1 do
    if desktopView.modules[i] = Self then
      begin
        index := i;
        Break;
      end;
  if index > -1 then
  begin
    removeFromSelectedModuleArray();
    if index + 1 < c then // if this wasn't the last item
      for i := index to c - 2 do // shift all items
        desktopView.modules[i] := desktopView.modules[i + 1];
    SetLength(desktopView.modules, c - 1);
  end;
end;

procedure TModule.removeFromSelectedModuleArray();
var
  desktopView: TDesktopView;
  i, c, index: Integer;
begin
  // detect the position in the module array of the module that will be deleted
  desktopView := TDesktopView(Parent);
  c := Length(desktopView.selectedModules);
  index := -1;
  for i := 0 to c - 1 do
    if desktopView.selectedModules[i] = Self then
      begin
        index := i;
        Break;
      end;
  if index > -1 then
  begin
    if index + 1 < c then // if this wasn't the last item
      for i := index to c - 2 do // shift all items
        desktopView.selectedModules[i] := desktopView.selectedModules[i + 1];
    SetLength(desktopView.selectedModules, c - 1);
  end;
end;

procedure TModule.closeDetails(freeDetails: Boolean);
var
  i, position: Integer;
begin
  position := -1;
  for i := 0 to Length(formLightning.formDetails) - 1 do
    if formLightning.formDetails[i].module = Self then
    begin
      position := i;
      Break;
    end;
  if position > -1 then
  begin
    if freeDetails then
      formLightning.formDetails[position].Free();
    // shift all other details windows in the array
    for i := position to Length(formLightning.formDetails) - 2 do
      formLightning.formDetails[i] := formLightning.formDetails[i + 1];
    SetLength(formLightning.formDetails, Length(formLightning.formDetails) - 1);
  end;
end;

procedure TModule.refreshOutputConnected();
var
  i: Integer;
begin
  // refresh this module's outputConnected list
  outputConnected := '';
  // make a string showing the connection state of each output
  for i := 1 to module^.outputs do
    if output[i].isConnected() then
      outputConnected := outputConnected + '1|'
    else
      outputConnected := outputConnected + '|';
end;

procedure TModule.refreshInputConnected();
var
  i: Integer;
begin
  // refresh this module's inputConnected list
  inputConnected := '';
  for i := 1 to module^.inputs do // make a string showing the connection state of each input
    if input[i].isConnected() then
      inputConnected := inputConnected + '1|'
    else
      inputConnected := inputConnected + '|';
end;

procedure TModule.preventCircle(checkModule: TModule);
var
  i: Integer;
begin
  for i := 1 to module^.outputs do
    if output[i].nextModule = checkModule then
    begin
      // circle was found, disconnect module
      output[i].disconnect();
    end
    else
    begin
      // no circle found yet, check connected modules
      if output[i].isConnected() then
        TModule(output[i].nextModule).preventCircle(checkModule);
    end;
end;

procedure TModule.moduleMouseMove(Sender: TObject; Shift: TShiftState; X, Y: Integer);
begin
  TDesktopView(Parent).mouseOut();
end;

procedure TModule.moduleMouseDown(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
var
  bPoint: TPoint;
begin
  if not(lightningSettings.lock) then
  begin
    if (Sender is TModuleColor) or (Sender is TModuleLayer) then
      Sender := TControl(Sender).Parent;
    case Button of
      mbLeft:
        if ssCtrl in Shift then
          TModule(Sender).setIsSelected(not(TModule(Sender).getIsSelected()))
        else
        begin
          // calculate the offset of the cursor relative to the dragged control
          GetCursorPos(bPoint);
          bPoint := TControl(Sender).ScreenToClient(bPoint);
          TDesktopView(TModule(Sender).Parent).deltaX := bPoint.X;
          TDesktopView(TModule(Sender).Parent).deltaY := bPoint.Y;
          TModule(Sender).BeginDrag(True);
        end;
    end;
  end;
end;

procedure TModule.moduleMouseUp(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
var
  bPoint: TPoint;
  i: Integer;
  desktopView: TDesktopView;
begin
  if not(lightningSettings.lock) then
    case Button of
      mbRight:
      begin
        // show context menu
        if (Sender is TModuleColor) or (Sender is TModuleLayer) then
          Sender := TControl(Sender).Parent;
        desktopView := TDesktopView(Parent);
        desktopView.contextModule := TModule(Sender);
        if not(desktopView.contextModule.getIsSelected()) then
        begin
          desktopView.allModulesSetIsSelected(False);
          desktopView.contextModule.setIsSelected(True);
        end;
        formLightning.miContextExcludeModuleFromFadeTime.Visible        := @desktopView.contextModule.openscene <> nil;
        formLightning.miContextApplytoAllScenesinCurrentProgram.Visible := @desktopView.contextModule.openscene <> nil;
        formLightning.miContextExcludeModuleFromFadeTime.Checked        := desktopView.contextModule.excludeFromFadeTime;
        for i := 0 to 9 do
          formLightning.menuItemContextSetLayer.Items[i].Checked := False;
        for i := 0 to Length(TDesktopView(Parent).selectedModules) - 1 do
          formLightning.menuItemContextSetLayer.Items[desktopView.selectedModules[i].getLayer()].Checked := True;
        GetCursorPos(bPoint);
        formLightning.pmContext.Popup(bPoint.X, bPoint.Y);
      end;
    end;
end;

procedure TModule.moduleStartDrag(Sender: TObject; var DragObject: TDragObject);
var
  bPoint: TPoint;
  bRect:  TRect;
  selTop, selLeft, selWidth, selHeight: Integer;
begin
  if not(getIsSelected()) then
  begin
    // if the module is unselected, deselect all other modules too.
    TDesktopView(TModule(Sender).Parent).allModulesSetIsSelected(False);
  end;
  // calculate and apply the cursor clipping area,
  // so that the cursor will be restricted to the client area of the main window.
  if not(TDesktopView(TModule(Sender).Parent).selectedModulesGetBounds(selTop, selLeft, selWidth, selHeight)) then
  begin
    selTop := TControl(Sender).Top;
    selLeft := TControl(Sender).Left;
    selWidth := TControl(Sender).Width;
    selHeight := TControl(Sender).Height;
  end;
  bPoint       := TDesktopView(TModule(Sender).Parent).ClientToScreen(Point(0, 0));
  bRect.Top    := bPoint.Y + (TControl(Sender).Top - selTop) + TDesktopView(TModule(Sender).Parent).deltaY;
  bRect.Left   := bPoint.X + (TControl(Sender).Left - selLeft) + TDesktopView(TModule(Sender).Parent).deltaX;
  bRect.Right  := bPoint.X + TDesktopView(TModule(Sender).Parent).ClientWidth;
  bRect.Bottom := bPoint.Y + TDesktopView(TModule(Sender).Parent).ClientHeight;
  ClipCursor(@bRect);
  // disable the dragged control to prevent any accidental actions on it
  TControl(Sender).Enabled := False;
end;

procedure TModule.moduleEndDrag(Sender, Target: TObject; X, Y: Integer);
var
  i: Integer;
  desktopView: TDesktopView;
begin
  desktopView := TDesktopView(Parent);
  // if the selection consists of newly created controls that are
  // not visible yet, show them.
  if not(TModule(Sender).Visible) then
  begin
    // show the dragged control
    TModule(Sender).Visible := True;
    // show any other selected controls
    for i := 0 to Length(desktopView.selectedModules) - 1 do
      if desktopView.selectedModules[i] <> TModule(Sender) then
        desktopView.selectedModules[i].Visible := True;
  end;
  // reenable the control
  TControl(Sender).Enabled := True;
  desktopView.clearFocusRect();
  desktopView.Repaint();
  // cancel the cursor clipping
  ClipCursor(nil);
end;

procedure TModule.lInfoClick(Sender: TObject);
var
  newCaption: string;
begin
  newCaption := getCaption();
  if InputQuery('Module Title', 'Please enter a new title for the selected module:', newCaption) then
  begin
    if newCaption = '' then
      newCaption := module^.title;
    setCaption(newCaption);
    // resize module width to let the new title fit
    updateWidth();
  end;
end;

procedure TModule.iUsedForSceneMouseDown(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
  if getIsSelected() then // check if the module is selected
  begin
    // module is selected, change state of all selected modules
    TDesktopView(Parent).selectedModulesSetIsUsedForScene(not(getIsUsedForScene()));
  end
  else
  begin
    // change state of only this module
    setIsUsedForScene(not(getIsUsedForScene()));
  end;
end;

function  TModule.toXML(): string;
var
  i: Integer;
  settings, xmlLayer, xmlSettings, xmlFadeTime, xmlColor, xmlTitle, xmlOutput: string;
begin
  SetLength(settings, maxLength);
  SetLength(settings, save(module, PChar(settings), Length(settings)));
  if settings <> '' then
    xmlSettings := '<settings>' + settings + '</settings>'
  else
    xmlSettings := '';
  if not excludeFromFadeTime then
    xmlFadeTime := '<fadetime />'
  else
    xmlFadeTime := '';
  // if the user has set a non-standard layer, save it
  if layer > 0 then
    xmlLayer := ' layer="' + IntToStr(layer) + '"'
  else
    xmlLayer := '';
  // if the user has set a non-standard module color, save it
  if ColorToRGB(getColor()) <> ColorToRGB(clBtnFace) then
  begin
    xmlColor :=
     '<color>#' +
     IntToHex(GetRValue(ColorToRGB(getColor())), 2) +
     IntToHex(GetGValue(ColorToRGB(getColor())), 2) +
     IntToHex(GetBValue(ColorToRGB(getColor())), 2) +
     '</color>';
  end
  else
    xmlColor := '';
  // if the user has set a non-standard module title, save it
  if module^.title <> lInfo.Caption then
    xmlTitle := '<title>' + lInfo.Caption + '</title>'
  else
    xmlTitle := '';
  Result :=
    '<module id="' + sceneid + '" top="' + IntToStr(top) + '" left="' + IntToStr(left) + '"' + xmlLayer + '>'#13#10 +
    '<file>' + ext + '</file>' +
    xmlTitle +
    xmlColor +
    xmlFadeTime +
    xmlSettings + #13#10;
  for i := 1 to module^.outputs do
  begin
    xmlOutput := output[i].toXML();
    if xmlOutput <> '' then
      Result := Result + xmlOutput + #13#10;
  end;
  Result := Result + '</module>';
end;

function  TModule.toString(includeId: Boolean): string;
var
  strSettings, strColor, strTitle, strId, strLayer: string;
begin
  SetLength(strSettings, maxLength);
  SetLength(strSettings, save(module, PChar(strSettings), Length(strSettings)));
  // if the user has set a non-standard module color, save it
  if ColorToRGB(getColor()) <> ColorToRGB(clBtnFace) then
    strColor := IntToStr(ColorToRGB(getColor()))
  else
    strColor := '';
  // if the user has set a non-standard module title, save it
  if module^.title <> getCaption() then
    strTitle := getCaption()
  else
    strTitle := '';
  if includeId then
    strId := sceneid
  else
    strId := '';
  if layer > 0 then
    strLayer := IntToStr(layer)
  else
    strLayer := '';
  Result := ext + '|' + IntToStr(Top) + '|' + IntToStr(Left) + '|' + strId + '|' + booleanToString(excludeFromFadeTime) + '|' + strLayer + '|' + strColor + ',' + strTitle + '||' + strSettings;
end;

function TModule.getColor(): TColor;
begin
  Result := sColor.Brush.Color;
end;

procedure TModule.setColor(color: TColor);
begin
  sColor.Brush.Color := color;
  sColor.Pen.Color   := color;
end;

function TModule.getCaption(): string;
begin
  Result := lInfo.Caption;
end;

procedure TModule.setCaption(c: string);
begin
  lInfo.Caption := c;
end;

procedure TModule.updateWidth();
var
  i, largest, minLeft, minWidth: Integer;
begin
  // resize module width based on label width
  lInfo.AutoSize := True;
  lInfo.Alignment := taLeftJustify;
  if module^.width > -1 then
    Width := module^.width
  else
    Width := 0;
  // determine whether the inputs or outputs need most space
  // and use the corresponding amount to determine the minimum required width
  if module^.inputs > module^.outputs then
    largest := module^.inputs
  else
    largest := module^.outputs;
  minWidth := 8 + largest * 24 + 1;
  if minWidth > Width then
    Width := minWidth;
  // determine the minimum width the control needs to allow for the info label,
  // which needs to be a multiply of 8. a margin of 8 pixels is applied.
  minWidth := 2 * 8 + Ceil(lInfo.Width / 8) * 8;
  if Width < minWidth then
    Width := minWidth;
  with lInfo do
  begin
    AutoSize  := False;
    Left      := 1;
    Width     := Self.Width - 2 * Left;
    Alignment := taCenter;
  end;
  // reposition outputs based on module width
  if module^.outputs > 0 then
  begin
    minLeft := Width - module^.outputs * 24 - 1;
    for i := 1 to module^.outputs do
      output[i].Left := minLeft + (i - 1) * 24;
  end;
  // reposition "usedforscene" button
  if iUsedForScene <> nil then
    iUsedForScene.Left := Width - iUsedForScene.Width - 1;
  // resize colored "title bar"
  sColor.Width := Width - 2;
  // reposition "settings" button
  if iSettings <> nil then
    iSettings.Left := Self.Width - 1 - 10;
end;

function TModule.getIsUsedForScene(): Boolean;
begin
  Result := usedForScene;
end;

procedure TModule.setIsUsedForScene(usedForScene: Boolean);
begin
  // check if the module is scene compatible
  if @openscene <> nil then
  begin
    Self.usedForScene := usedForScene;
    if usedForScene then
      iUsedForScene.Picture.Bitmap.LoadFromResourceName(hInstance, 'usedforscenedown')
    else
      iUsedForScene.Picture.Bitmap.LoadFromResourceName(hInstance, 'usedforsceneup');
  end;
end;

procedure TModule.setIsSelected(selected: Boolean);
begin
  if selected then
  begin
    BevelOuter := bvLowered;
    addToSelectedModuleArray();
  end
  else
  begin
    BevelOuter := bvRaised;
    removeFromSelectedModuleArray();
  end;
end;

// Result = selected?
function  TModule.getIsSelected(): Boolean;
begin
  Result := BevelOuter = bvLowered;
end;

procedure TModule.updateLayer();
begin
  Visible := lightningSettings.showLayer[layer];
  if layer > 0 then
  begin
    // module on layer > 0
    labelLayer.Caption := IntToStr(layer);
    lInfo.Font.Color := clGray;
  end
  else
  begin
    // module on layer 0
    labelLayer.Caption := '';
    lInfo.Font.Color := $00533200;
  end;
  if not(Visible) then
    setIsSelected(False);
end;

procedure TModule.setLayer(layer: Integer);
begin
  Self.layer := layer;
  updateLayer();
end;

// Result = layer
function  TModule.getLayer(): Integer;
begin
  Result := layer;
end;

procedure TModule.setLocked(locked: Boolean);
begin
  if locked then
  begin
    // if lock is enabled, module titles can't be changed.
    lInfo.Cursor  := crDefault;
    lInfo.OnClick := nil;
  end
  else
  begin
    // if lock is disabled, a handpoint cursor appears and
    // module titles can be changed.
    lInfo.Cursor  := crHandPoint;
    lInfo.OnClick := lInfoClick;
  end;
end;

procedure TModule.BringToFront();
var
  desktopView: TDesktopView;
  i, j, c: Integer;
begin
  inherited;
  desktopView := TDesktopView(Parent);
  c := Length(desktopView.modules);
  // find module in modules array
  for i := 0 to c - 1 do
    if desktopView.modules[i] = Self then
    begin
      // shift all modules after the module so that they replace the module
      for j := i to c - 2 do
        desktopView.modules[j] := desktopView.modules[j + 1];
      // store a reference to the module at the end of the array
      desktopView.modules[c - 1] := Self;
      Break;
    end;
end;

procedure TModule.SendToBack();
var
  desktopView: TDesktopView;
  i, j, c: Integer;
begin
  inherited;
  desktopView := TDesktopView(Parent);
  c := Length(desktopView.modules);
  // find context module in modules array
  for i := 0 to c - 1 do
    if desktopView.modules[i] = Self then
    begin
      // shift all modules before the context module so that they replace the context module
      for j := i - 1 downto 0 do
        desktopView.modules[j + 1] := desktopView.modules[j];
      // store a reference to the context module at the beginning of the array
      desktopView.modules[0] := Self;
      Break;
    end;
end;

// processes data output on a module's output (used by TModule.actionInput() and by component modules)
procedure TModule.actionOutput(input, output: Integer; bridge: TSpectrum);
var
  level: TLevel;
begin
//  if formTrace <> nil then
//    formTrace.add('actionOutput(' + ext + ',' + IntToStr(input) + ',' + IntToStr(output) + ')');
  if action(module, input, output, PChar(inputConnected), PChar(outputConnected), bridge) then
  begin
    if not(minimized) and lightningSettings.visualiseDataFlow and not(TDesktopView(Parent).Parent is TModule) then
    begin
      level := spectrumAverage(bridge, fftMin, fftMax);
      Self.output[output].setLevel(level);
      if Self.output[output].isConnected() then
      begin
        TModule(Self.output[output].nextModule).input[Self.output[output].nextInput].setLevel(level);
      end;
    end;
    if Self.output[output].isConnected() then
    begin
      // call the next module in the chain
      TModule(Self.output[output].nextModule).actionInput(Self.output[output].nextInput, bridge, True);
    end;
  end;
end;

// send data to unconnected inputs of LUA modules
procedure TModule.actionUnconnectedInputs(input: Integer);
var
  bString: string;
  i, lastInputConnectedBeforeThisInput, lastInputConnected: Integer;
  dummy: TSpectrum;
begin
  // TODO consider moving this piece of LUA specific code into
  // action() function of unitmodulelua.pas.
  // for LUA modules, Lightning has to garuantee that a signal is sent,
  // even to unconnected inputs.
  // find unconnected inputs inbetween this input and
  // the previous connected input and call action on them.
  bString := inputConnected;
  i := 1;
  lastInputConnectedBeforeThisInput := 0;
  lastInputConnected := 1;
  repeat
    if parse(bString, '|') <> '' then
    begin
      if i < input then
        lastInputConnectedBeforeThisInput := i;
      lastInputConnected := i;
    end;
    Inc(i);
  until bString = '';
  for i := lastInputConnectedBeforeThisInput + 1 to input - 1 do
  begin
    createSpectrum(dummy, levelMin);
    actionInput(i, dummy, False);
  end;
  // if this is the last connected input,
  // find unconnected inputs after this input and process them.
  if lastInputConnected = input then
  begin
    for i := input + 1 to module^.inputs do
    begin
      createSpectrum(dummy, levelMin);
      actionInput(i, dummy, False);
    end;
  end;
end;

// process data arriving on a module's input
procedure TModule.actionInput(input: Integer; var bridge: TSpectrum; checkLUA: Boolean);
var
  i: Integer;
begin
//  if formTrace <> nil then
//    formTrace.add('actionInput(' + ext + ',' + IntToStr(input) + ')');
  // check if this is a Processing or Output module and is a LUA moduel
  if (kind = mkLUA)
     and
     checkLUA
     and
     (module^.inputs > 0)
     then
  begin
    actionUnconnectedInputs(input);
  end;
  if module^.outputs > 0 then
  begin
    // the module has output connectors, so it is an Input or Processing module.
    // Input modules are always at the beginning of the chain.
    // Processing modules are anywhere in the middle of the chain.
    for i := 1 to module^.outputs do // and for each output allow the module to act
      actionOutput(input, i, bridge);
  end
  else
  begin
    // the module has no outputs, so it is an Output module.
    // Output modules are always at the end of the chain.
    action(module, input, 0, PChar(inputConnected), '', bridge);
  end;
end;

end.
