unit unitmodulecomponent;

(* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1
 *
 * The contents of this file are subject to the Mozilla Public License Version 
 * 1.1 (the "License"); you may not use this file except in compliance with 
 * the License. You may obtain a copy of the License at 
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Lightning component support code.
 *
 * The Initial Developer of the Original Code is
 * Tomsoft.
 * Portions created by Tomsoft are Copyright (C) 2009
 * Tomsoft. All Rights Reserved.
 *
 * Contributor(s):
 *   Tommy Sools <tom@sools.com>
 *
 * ***** END LICENSE BLOCK ***** *)

 interface

uses
  Windows,
  SysUtils,
  common,
  unitdesktopview;

const
  channels = 32;

//function  init(name: PChar; nameLength: DWORD): DWORD; cdecl;
//procedure finish(); cdecl;
function  create(controlHandle: THandle; fileName: string; desktopView: TDesktopView): PLightningModule; cdecl;
procedure destroy(lightningModule: PLightningModule); cdecl;
function  action(lightningModule: PLightningModule; input, output: Integer; inputConnected, outputConnected: PChar; var bridge: TSpectrum): Boolean; cdecl;
procedure show(lightningModule: PLightningModule; handle: THandle); cdecl;
procedure hide(lightningModule: PLightningModule); cdecl;
procedure open(lightningModule: PLightningModule; settings: PChar); cdecl;
function  save(lightningModule: PLightningModule; settings: PChar; settingsLength: DWORD): DWORD; cdecl;

implementation

uses
  unitmodule;

{
function  init(name: PChar; nameLength: DWORD): DWORD; cdecl;
begin
  Result := return('', name, nameLength);
end;

procedure finish; cdecl;
begin
  //
end;
}

function  create(controlHandle: THandle; fileName: string; desktopView: TDesktopView): PLightningModule; cdecl;
var
  moduleIn: TModule;
  moduleOut: TModule;
  inCount: Integer;
  outCount: Integer;
  i: Integer;
  bString: string;
begin
  // locate the "in_component" module inside this component's desktopview
  moduleIn := nil;
  for i := 0 to Length(desktopView.modules) - 1 do
    if desktopView.modules[i].kind = mkComponentIn then
    begin
      moduleIn := desktopView.modules[i];
      Break;
    end;
  // count connections starting at "in_component" module
  inCount := 0;
  if moduleIn <> nil then
    for i := moduleIn.module^.outputs downto 1 do
      if moduleIn.output[i].isConnected() then
      begin
        inCount := i;
        Break;
      end;
  // locate the "out_component" module inside this component's desktopview
  moduleOut := nil;
  for i := 0 to Length(desktopView.modules) - 1 do
    if desktopView.modules[i].kind = mkComponentOut then
    begin
      moduleOut := desktopView.modules[i];
      Break;
    end;
  // count connections ending at "out_component" module
  outCount := 0;
  if moduleOut <> nil then
    for i := moduleOut.module^.inputs downto 1 do
      if moduleOut.input[i].isConnected() then
      begin
        outCount := i;
        Break;
      end;

  New(Result);
  with Result^ do
  begin
    handle        := controlHandle;
    bString       := ExtractFileName(fileName);
    GetMem(title, {maxLength}Length(bString) + 1);
    StrPCopy(title, {''}bString);
    settings      := False;
    width         := -1;
    height        := -1;
    inputTop      := -1;
    infoTop       := -1;
    outputTop     := -1;
    inputs        := inCount;
    outputs       := outCount;
    GetMem(inputTooltip, maxLength);
    StrCopy(inputTooltip, '');
    for i := 1 to inputs do
      inputTooltip := StrCat(inputTooltip, PChar('Channel ' + IntToStr(i) + ' spectrum|'));
    GetMem(outputTooltip, maxLength);
    StrCopy(outputTooltip, '');
    for i := 1 to outputs do
      outputTooltip := StrCat(outputTooltip, PChar('Channel ' + IntToStr(i) + ' spectrum|'));
    details       := Pointer(desktopView);
    reserved      := nil;
  end;
end;

procedure destroy(lightningModule: PLightningModule); cdecl;
begin
  TDesktopView(lightningModule^.details).Free();
  FreeMem(lightningModule^.outputTooltip);
  FreeMem(lightningModule^.inputTooltip);
  FreeMem(lightningModule^.title);
  Dispose(lightningModule);
end;

function  action(lightningModule: PLightningModule; input, output: Integer; inputConnected, outputConnected: PChar; var bridge: TSpectrum): Boolean; cdecl;
var
  i: Integer;
  desktopView: TDesktopView;
  module: TModule;
begin
  if input > 0 then
  begin
    // the data originates from normal modules
    // locate the "in_component" module inside this component's desktopview
    desktopView := TDesktopView(lightningModule^.details);
    module := nil;
    for i := 0 to Length(desktopView.modules) - 1 do
      if desktopView.modules[i].kind = mkComponentIn then
      begin
        module := desktopView.modules[i];
        Break;
      end;
    if (module <> nil) and (input = output) then
    begin
      // modify the output chain by breaking it and calling action on
      // a corresponding output of the in_component module.
      module.actionOutput(0, output, bridge);
    end;
    Result := False;
  end
  else
  begin
    // the data originates from the out_component module, so simply forward it
    // to the next module in the chain.
    Result := True;
  end;
end;

procedure show(lightningModule: PLightningModule; handle: THandle); cdecl;
begin
  //
end;

procedure hide(lightningModule: PLightningModule); cdecl;
begin
  //
end;

procedure open(lightningModule: PLightningModule; settings: PChar); cdecl;
begin
  //
end;

function  save(lightningModule: PLightningModule; settings: PChar; settingsLength: DWORD): DWORD; cdecl;
begin
  Result := return('', settings, settingsLength);
end;

end.
